import os
import sys
from unittest.mock import call, MagicMock

import pytest
import torch
import torch.nn as nn
from torch.utils.data.distributed import DistributedSampler

import ignite.distributed as idist
import ignite.handlers as handlers
from ignite.contrib.engines.common import (
    _setup_logging,
    add_early_stopping_by_val_score,
    gen_save_best_models_by_val_score,
    save_best_model_by_val_score,
    setup_any_logging,
    setup_clearml_logging,
    setup_common_training_handlers,
    setup_mlflow_logging,
    setup_neptune_logging,
    setup_plx_logging,
    setup_tb_logging,
    setup_trains_logging,
    setup_visdom_logging,
    setup_wandb_logging,
)
from ignite.engine import Engine, Events
from ignite.handlers import DiskSaver, TerminateOnNan


class DummyModel(nn.Module):
    def __init__(self):
        super(DummyModel, self).__init__()
        self.net = nn.Linear(1, 1)

    def forward(self, x):
        return self.net(x)


def _test_setup_common_training_handlers(
    dirname,
    device,
    rank=0,
    local_rank=0,
    distributed=False,
    lr_scheduler=None,
    save_handler=None,
    output_transform=lambda loss: loss,
):
    lr = 0.01
    step_size = 100
    gamma = 0.5
    num_iters = 100
    num_epochs = 10

    model = DummyModel().to(device)
    if distributed and "cuda" in torch.device(device).type:
        model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[local_rank], output_device=local_rank)
    optimizer = torch.optim.SGD(model.parameters(), lr=lr)

    if lr_scheduler is None:
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma)
    elif isinstance(lr_scheduler, str) and lr_scheduler == "ignite|LRScheduler":
        from ignite.handlers import LRScheduler

        lr_scheduler = LRScheduler(torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma))
    elif isinstance(lr_scheduler, str) and lr_scheduler == "ignite":
        from ignite.handlers import PiecewiseLinear

        milestones_values = [(0, 0.0), (step_size, lr), (num_iters * (num_epochs - 1), 0.0)]
        lr_scheduler = PiecewiseLinear(optimizer, param_name="lr", milestones_values=milestones_values)
    else:
        raise ValueError(f"Unknown lr_scheduler: {lr_scheduler}")

    def update_fn(engine, batch):
        optimizer.zero_grad()
        x = torch.tensor([batch], requires_grad=True, device=device)
        y_pred = model(x)
        loss = y_pred.mean()
        loss.backward()
        optimizer.step()
        return output_transform(loss)

    train_sampler = None
    if distributed and idist.get_world_size() > 1:
        train_sampler = MagicMock(spec=DistributedSampler)
        train_sampler.set_epoch = MagicMock()

    trainer = Engine(update_fn)
    setup_common_training_handlers(
        trainer,
        train_sampler=train_sampler,
        to_save={"model": model, "optimizer": optimizer},
        save_every_iters=75,
        output_path=dirname,
        save_handler=save_handler,
        lr_scheduler=lr_scheduler,
        with_gpu_stats=False,
        output_names=["batch_loss"],
        with_pbars=True,
        with_pbar_on_iters=True,
        log_every_iters=50,
    )

    data = [i * 0.1 for i in range(num_iters)]
    trainer.run(data, max_epochs=num_epochs)

    # check handlers
    handlers = trainer._event_handlers[Events.ITERATION_COMPLETED]
    for cls in [
        TerminateOnNan,
    ]:
        assert any([isinstance(h[0], cls) for h in handlers]), f"{handlers}"
    assert "batch_loss" in trainer.state.metrics

    # Check saved checkpoint
    if rank == 0:
        if save_handler is not None:
            dirname = save_handler.dirname
        checkpoints = list(os.listdir(dirname))
        assert len(checkpoints) == 1
        for v in [
            "training_checkpoint",
        ]:
            assert any([v in c for c in checkpoints])

    # Check LR scheduling
    assert optimizer.param_groups[0]["lr"] <= lr * gamma ** (
        (num_iters * num_epochs - 1) // step_size
    ), f"{optimizer.param_groups[0]['lr']} vs {lr * gamma ** ((num_iters * num_epochs - 1) // step_size)}"


def test_asserts_setup_common_training_handlers():
    trainer = Engine(lambda e, b: None)

    with pytest.raises(
        ValueError,
        match=r"If to_save argument is provided then output_path or save_handler arguments should be also defined",
    ):
        setup_common_training_handlers(trainer, to_save={})

    with pytest.raises(ValueError, match=r"Arguments output_path and save_handler are mutually exclusive"):
        setup_common_training_handlers(trainer, to_save={}, output_path="abc", save_handler=lambda c, f, m: None)

    with pytest.warns(UserWarning, match=r"Argument train_sampler is a distributed sampler"):
        train_sampler = MagicMock(spec=DistributedSampler)
        setup_common_training_handlers(trainer, train_sampler=train_sampler)

    if not torch.cuda.is_available():
        with pytest.raises(RuntimeError, match=r"This contrib module requires available GPU"):
            setup_common_training_handlers(trainer, with_gpu_stats=True)

    with pytest.raises(TypeError, match=r"Unhandled type of update_function's output."):
        trainer = Engine(lambda e, b: None)
        setup_common_training_handlers(
            trainer,
            output_names=["loss"],
            with_pbar_on_iters=False,
            with_pbars=False,
            with_gpu_stats=False,
            stop_on_nan=False,
            clear_cuda_cache=False,
        )
        trainer.run([1])


def test_no_warning_with_train_sampler(recwarn):
    from torch.utils.data import RandomSampler

    trainer = Engine(lambda e, b: None)
    train_sampler = RandomSampler([0, 1, 2])
    setup_common_training_handlers(trainer, train_sampler=train_sampler)
    assert len(recwarn) == 0, recwarn.pop()


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("WORLD_SIZE" not in os.environ, reason="Should have more than 1 worker")
def test_assert_setup_common_training_handlers_wrong_train_sampler(distributed_context_single_node_gloo):
    trainer = Engine(lambda e, b: None)

    from torch.utils.data.sampler import RandomSampler

    with pytest.raises(TypeError, match=r"Train sampler should be torch DistributedSampler"):
        train_sampler = RandomSampler([0, 1, 2, 3])
        setup_common_training_handlers(trainer, train_sampler)


def test_setup_common_training_handlers(dirname, capsys):
    _test_setup_common_training_handlers(dirname, device="cpu")

    # Check epoch-wise pbar
    captured = capsys.readouterr()
    out = captured.err.split("\r")
    out = list(map(lambda x: x.strip(), out))
    out = list(filter(None, out))
    assert "Epoch" in out[-1] or "Epoch" in out[-2], f"{out[-2]}, {out[-1]}"

    _test_setup_common_training_handlers(dirname, device="cpu", output_transform=lambda loss: [loss])

    # Check epoch-wise pbar
    captured = capsys.readouterr()
    out = captured.err.split("\r")
    out = list(map(lambda x: x.strip(), out))
    out = list(filter(None, out))
    assert "Epoch" in out[-1] or "Epoch" in out[-2], f"{out[-2]}, {out[-1]}"

    _test_setup_common_training_handlers(dirname, device="cpu", output_transform=lambda loss: {"batch_loss": loss})

    # Check epoch-wise pbar
    captured = capsys.readouterr()
    out = captured.err.split("\r")
    out = list(map(lambda x: x.strip(), out))
    out = list(filter(None, out))
    assert "Epoch" in out[-1] or "Epoch" in out[-2], f"{out[-2]}, {out[-1]}"


def test_setup_common_training_handlers_using_save_handler(dirname, capsys):
    save_handler = DiskSaver(dirname=dirname, require_empty=False)
    _test_setup_common_training_handlers(dirname=None, device="cpu", save_handler=save_handler)

    # Check epoch-wise pbar
    captured = capsys.readouterr()
    out = captured.err.split("\r")
    out = list(map(lambda x: x.strip(), out))
    out = list(filter(None, out))
    assert "Epoch" in out[-1] or "Epoch" in out[-2], f"{out[-2]}, {out[-1]}"


def test_save_best_model_by_val_score(dirname):
    acc_scores = [0.1, 0.2, 0.3, 0.4, 0.3, 0.5, 0.6, 0.61, 0.7, 0.5]

    def setup_trainer():
        trainer = Engine(lambda e, b: None)
        evaluator = Engine(lambda e, b: None)
        model = DummyModel()

        @trainer.on(Events.EPOCH_COMPLETED)
        def validate(engine):
            evaluator.run([0, 1])

        @evaluator.on(Events.EPOCH_COMPLETED)
        def set_eval_metric(engine):
            acc = acc_scores[trainer.state.epoch - 1]
            engine.state.metrics = {"acc": acc, "loss": 1 - acc}

        return trainer, evaluator, model

    trainer, evaluator, model = setup_trainer()

    save_best_model_by_val_score(dirname, evaluator, model, metric_name="acc", n_saved=2, trainer=trainer)

    trainer.run([0, 1], max_epochs=len(acc_scores))

    assert set(os.listdir(dirname)) == {"best_model_8_val_acc=0.6100.pt", "best_model_9_val_acc=0.7000.pt"}

    for fname in os.listdir(dirname):
        os.unlink(f"{dirname}/{fname}")

    trainer, evaluator, model = setup_trainer()

    save_best_model_by_val_score(
        dirname, evaluator, model, metric_name="loss", n_saved=2, trainer=trainer, score_sign=-1.0
    )

    trainer.run([0, 1], max_epochs=len(acc_scores))

    assert set(os.listdir(dirname)) == {"best_model_8_val_loss=-0.3900.pt", "best_model_9_val_loss=-0.3000.pt"}


def test_gen_save_best_models_by_val_score():
    acc_scores = [0.1, 0.2, 0.3, 0.4, 0.3, 0.5, 0.6, 0.61, 0.7, 0.5]
    loss_scores = [0.9, 0.8, 0.7, 0.6, 0.7, 0.5, 0.4, 0.39, 0.3, 0.5]

    def setup_trainer():
        trainer = Engine(lambda e, b: None)
        evaluator = Engine(lambda e, b: None)
        model = DummyModel()

        @trainer.on(Events.EPOCH_COMPLETED)
        def validate(engine):
            evaluator.run([0, 1])

        @evaluator.on(Events.EPOCH_COMPLETED)
        def set_eval_metric(engine):
            acc = acc_scores[trainer.state.epoch - 1]
            loss = loss_scores[trainer.state.epoch - 1]
            engine.state.metrics = {"acc": acc, "loss": loss}

        return trainer, evaluator, model

    trainer, evaluator, model = setup_trainer()

    save_handler = MagicMock()

    gen_save_best_models_by_val_score(
        save_handler, evaluator, {"a": model, "b": model}, metric_name="acc", n_saved=2, trainer=trainer
    )

    trainer.run([0, 1], max_epochs=len(acc_scores))

    assert save_handler.call_count == len(acc_scores) - 2  # 2 score values (0.3 and 0.5) are not the best
    obj_to_save = {"a": model.state_dict(), "b": model.state_dict()}
    save_handler.assert_has_calls(
        [
            call(
                obj_to_save,
                f"best_checkpoint_{e}_val_acc={p:.4f}.pt",
                dict([("basename", "best_checkpoint"), ("score_name", "val_acc"), ("priority", p)]),
            )
            for e, p in zip([1, 2, 3, 4, 6, 7, 8, 9], [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.61, 0.7])
        ],
        any_order=True,
    )

    trainer, evaluator, model = setup_trainer()

    save_handler = MagicMock()

    gen_save_best_models_by_val_score(
        save_handler,
        evaluator,
        {"a": model, "b": model},
        metric_name="loss",
        n_saved=2,
        trainer=trainer,
        score_sign=-1.0,
    )

    trainer.run([0, 1], max_epochs=len(acc_scores))

    assert save_handler.call_count == len(acc_scores) - 2  # 2 score values (-0.7 and -0.5) are not the best
    obj_to_save = {"a": model.state_dict(), "b": model.state_dict()}
    save_handler.assert_has_calls(
        [
            call(
                obj_to_save,
                f"best_checkpoint_{e}_val_loss={p:.4f}.pt",
                dict([("basename", "best_checkpoint"), ("score_name", "val_loss"), ("priority", p)]),
            )
            for e, p in zip([1, 2, 3, 4, 6, 7, 8, 9], [-0.9, -0.8, -0.7, -0.6, -0.5, -0.4, -0.39, -0.3])
        ],
        any_order=True,
    )


def test_add_early_stopping_by_val_score():
    acc_scores = [0.1, 0.2, 0.3, 0.4, 0.3, 0.3, 0.2, 0.1, 0.1, 0.0]

    def setup_trainer():
        trainer = Engine(lambda e, b: None)
        evaluator = Engine(lambda e, b: None)

        @trainer.on(Events.EPOCH_COMPLETED)
        def validate(engine):
            evaluator.run([0, 1])

        @evaluator.on(Events.EPOCH_COMPLETED)
        def set_eval_metric(engine):
            acc = acc_scores[trainer.state.epoch - 1]
            engine.state.metrics = {"acc": acc, "loss": 1 - acc}

        return trainer, evaluator

    trainer, evaluator = setup_trainer()

    add_early_stopping_by_val_score(patience=3, evaluator=evaluator, trainer=trainer, metric_name="acc")

    state = trainer.run([0, 1], max_epochs=len(acc_scores))

    assert state.epoch == 7

    trainer, evaluator = setup_trainer()

    add_early_stopping_by_val_score(
        patience=3, evaluator=evaluator, trainer=trainer, metric_name="loss", score_sign=-1.0
    )

    state = trainer.run([0, 1], max_epochs=len(acc_scores))

    assert state.epoch == 7


def test_deprecated_setup_any_logging():
    with pytest.raises(DeprecationWarning, match=r"deprecated since version 0.4.0"):
        setup_any_logging(None, None, None, None, None, None)


def test__setup_logging_wrong_args():
    with pytest.raises(TypeError, match=r"Argument optimizers should be either a single optimizer or"):
        _setup_logging(MagicMock(), MagicMock(), "abc", MagicMock(), 1)

    with pytest.raises(TypeError, match=r"Argument evaluators should be either a single engine or"):
        _setup_logging(MagicMock(), MagicMock(), MagicMock(spec=torch.optim.SGD), "abc", 1)


def _test_setup_logging(
    setup_logging_fn,
    kwargs_dict,
    output_handler_cls,
    opt_params_handler_cls,
    with_eval=True,
    with_optim=True,
    as_class=False,
    log_every_iters=1,
):
    trainer = Engine(lambda e, b: b)
    evaluators = None
    optimizers = None

    if with_eval:
        evaluator = Engine(lambda e, b: None)
        acc_scores = [0.1, 0.2, 0.3, 0.4, 0.3, 0.3, 0.2, 0.1, 0.1, 0.0]

        @trainer.on(Events.EPOCH_COMPLETED)
        def validate(engine):
            evaluator.run([0, 1])

        @evaluator.on(Events.EPOCH_COMPLETED)
        def set_eval_metric(engine):
            engine.state.metrics = {"acc": acc_scores[trainer.state.epoch - 1]}

        evaluators = {"validation": evaluator}
        if as_class:
            evaluators = evaluators["validation"]

    if with_optim:
        t = torch.tensor([0])
        optimizers = {"optimizer": torch.optim.SGD([t], lr=0.01)}
        if as_class:
            optimizers = optimizers["optimizer"]

    kwargs_dict["trainer"] = trainer
    kwargs_dict["optimizers"] = optimizers
    kwargs_dict["evaluators"] = evaluators
    kwargs_dict["log_every_iters"] = log_every_iters

    x_logger = setup_logging_fn(**kwargs_dict)

    handlers = trainer._event_handlers[Events.ITERATION_COMPLETED]
    for cls in [
        output_handler_cls,
    ]:
        assert any([isinstance(h[0], cls) for h in handlers]), f"{handlers}"

    if with_optim:
        handlers = trainer._event_handlers[Events.ITERATION_STARTED]
        for cls in [
            opt_params_handler_cls,
        ]:
            assert any([isinstance(h[0], cls) for h in handlers]), f"{handlers}"

    if with_eval:
        handlers = evaluator._event_handlers[Events.COMPLETED]
        for cls in [
            output_handler_cls,
        ]:
            assert any([isinstance(h[0], cls) for h in handlers]), f"{handlers}"

    data = [0, 1, 2]
    trainer.run(data, max_epochs=10)

    if "output_path" in kwargs_dict:
        tb_files = list(os.listdir(kwargs_dict["output_path"]))
        assert len(tb_files) == 1
        for v in [
            "events",
        ]:
            assert any([v in c for c in tb_files]), f"{tb_files}"

    return x_logger


def test_setup_tb_logging(dirname):
    tb_logger = _test_setup_logging(
        setup_logging_fn=setup_tb_logging,
        kwargs_dict={"output_path": dirname / "t1"},
        output_handler_cls=handlers.tensorboard_logger.OutputHandler,
        opt_params_handler_cls=handlers.tensorboard_logger.OptimizerParamsHandler,
        with_eval=False,
        with_optim=False,
    )
    tb_logger.close()
    tb_logger = _test_setup_logging(
        setup_logging_fn=setup_tb_logging,
        kwargs_dict={"output_path": dirname / "t2"},
        output_handler_cls=handlers.tensorboard_logger.OutputHandler,
        opt_params_handler_cls=handlers.tensorboard_logger.OptimizerParamsHandler,
        with_eval=True,
        with_optim=True,
    )
    tb_logger.close()
    tb_logger = _test_setup_logging(
        setup_logging_fn=setup_tb_logging,
        kwargs_dict={"output_path": dirname / "t3"},
        output_handler_cls=handlers.tensorboard_logger.OutputHandler,
        opt_params_handler_cls=handlers.tensorboard_logger.OptimizerParamsHandler,
        with_eval=True,
        with_optim=True,
        as_class=True,
        log_every_iters=None,
    )
    tb_logger.close()


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_setup_visdom_logging(visdom_offline_logfile):
    vis_logger = _test_setup_logging(
        setup_logging_fn=setup_visdom_logging,
        kwargs_dict={"offline": True, "log_to_filename": visdom_offline_logfile},
        output_handler_cls=handlers.visdom_logger.OutputHandler,
        opt_params_handler_cls=handlers.visdom_logger.OptimizerParamsHandler,
        with_eval=False,
        with_optim=False,
    )
    vis_logger.close()

    vis_logger = _test_setup_logging(
        setup_logging_fn=setup_visdom_logging,
        kwargs_dict={"offline": True, "log_to_filename": visdom_offline_logfile},
        output_handler_cls=handlers.visdom_logger.OutputHandler,
        opt_params_handler_cls=handlers.visdom_logger.OptimizerParamsHandler,
        with_eval=True,
        with_optim=True,
    )
    vis_logger.close()


def test_setup_plx_logging():
    os.environ["POLYAXON_NO_OP"] = "1"

    _test_setup_logging(
        setup_logging_fn=setup_plx_logging,
        kwargs_dict={},
        output_handler_cls=handlers.polyaxon_logger.OutputHandler,
        opt_params_handler_cls=handlers.polyaxon_logger.OptimizerParamsHandler,
        with_eval=False,
        with_optim=False,
    )
    _test_setup_logging(
        setup_logging_fn=setup_plx_logging,
        kwargs_dict={},
        output_handler_cls=handlers.polyaxon_logger.OutputHandler,
        opt_params_handler_cls=handlers.polyaxon_logger.OptimizerParamsHandler,
        with_eval=True,
        with_optim=True,
    )


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_setup_mlflow_logging(dirname):
    mlf_logger = _test_setup_logging(
        setup_logging_fn=setup_mlflow_logging,
        kwargs_dict={"tracking_uri": str(dirname / "p1")},
        output_handler_cls=handlers.mlflow_logger.OutputHandler,
        opt_params_handler_cls=handlers.mlflow_logger.OptimizerParamsHandler,
        with_eval=False,
        with_optim=False,
    )
    mlf_logger.close()
    mlf_logger = _test_setup_logging(
        setup_logging_fn=setup_mlflow_logging,
        kwargs_dict={"tracking_uri": str(dirname / "p2")},
        output_handler_cls=handlers.mlflow_logger.OutputHandler,
        opt_params_handler_cls=handlers.mlflow_logger.OptimizerParamsHandler,
        with_eval=True,
        with_optim=True,
    )
    mlf_logger.close()


def test_setup_wandb_logging(dirname):
    from unittest.mock import patch

    with patch("ignite.contrib.engines.common.WandBLogger") as _:
        setup_wandb_logging(MagicMock())


def test_setup_clearml_logging():
    handlers.clearml_logger.ClearMLLogger.set_bypass_mode(True)

    with pytest.warns(UserWarning, match=r"running in bypass mode"):
        clearml_logger = _test_setup_logging(
            setup_logging_fn=setup_clearml_logging,
            kwargs_dict={},
            output_handler_cls=handlers.clearml_logger.OutputHandler,
            opt_params_handler_cls=handlers.clearml_logger.OptimizerParamsHandler,
            with_eval=False,
            with_optim=False,
        )
        clearml_logger.close()
        clearml_logger = _test_setup_logging(
            setup_logging_fn=setup_clearml_logging,
            kwargs_dict={},
            output_handler_cls=handlers.clearml_logger.OutputHandler,
            opt_params_handler_cls=handlers.clearml_logger.OptimizerParamsHandler,
            with_eval=True,
            with_optim=True,
        )
        clearml_logger.close()
        clearml_logger = _test_setup_logging(
            setup_logging_fn=setup_trains_logging,
            kwargs_dict={},
            output_handler_cls=handlers.clearml_logger.OutputHandler,
            opt_params_handler_cls=handlers.clearml_logger.OptimizerParamsHandler,
            with_eval=True,
            with_optim=True,
        )
        clearml_logger.close()

    with pytest.warns(UserWarning, match="setup_trains_logging was renamed to setup_clearml_logging"):
        clearml_logger = _test_setup_logging(
            setup_logging_fn=setup_trains_logging,
            kwargs_dict={},
            output_handler_cls=handlers.clearml_logger.OutputHandler,
            opt_params_handler_cls=handlers.clearml_logger.OptimizerParamsHandler,
            with_eval=True,
            with_optim=True,
        )
        clearml_logger.close()


def test_setup_neptune_logging(dirname):
    npt_logger = _test_setup_logging(
        setup_logging_fn=setup_neptune_logging,
        kwargs_dict={"mode": "offline"},
        output_handler_cls=handlers.neptune_logger.OutputHandler,
        opt_params_handler_cls=handlers.neptune_logger.OptimizerParamsHandler,
        with_eval=False,
        with_optim=False,
    )
    npt_logger.close()
    npt_logger = _test_setup_logging(
        setup_logging_fn=setup_neptune_logging,
        kwargs_dict={"mode": "offline"},
        output_handler_cls=handlers.neptune_logger.OutputHandler,
        opt_params_handler_cls=handlers.neptune_logger.OptimizerParamsHandler,
        with_eval=True,
        with_optim=True,
    )
    npt_logger.close()


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(dirname, distributed_context_single_node_nccl):
    local_rank = distributed_context_single_node_nccl["local_rank"]
    device = idist.device()
    _test_setup_common_training_handlers(dirname, device, rank=local_rank, local_rank=local_rank, distributed=True)
    test_add_early_stopping_by_val_score()


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(dirname, distributed_context_single_node_gloo):
    device = idist.device()
    local_rank = distributed_context_single_node_gloo["local_rank"]
    _test_setup_common_training_handlers(dirname, device, rank=local_rank, local_rank=local_rank, distributed=True)
    _test_setup_common_training_handlers(
        dirname, device, rank=local_rank, local_rank=local_rank, distributed=True, lr_scheduler="ignite|LRScheduler"
    )
    _test_setup_common_training_handlers(
        dirname, device, rank=local_rank, local_rank=local_rank, distributed=True, lr_scheduler="ignite"
    )
    test_add_early_stopping_by_val_score()


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(dirname, distributed_context_multi_node_gloo):
    device = idist.device()
    rank = distributed_context_multi_node_gloo["rank"]
    _test_setup_common_training_handlers(dirname, device, rank=rank)
    test_add_early_stopping_by_val_score()


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(dirname, distributed_context_multi_node_nccl):
    local_rank = distributed_context_multi_node_nccl["local_rank"]
    rank = distributed_context_multi_node_nccl["rank"]
    device = idist.device()
    _test_setup_common_training_handlers(dirname, device, rank=rank, local_rank=local_rank, distributed=True)
    test_add_early_stopping_by_val_score()



import os

import pytest
import torch
from sklearn.metrics import r2_score

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.regression import R2Score


def test_zero_sample():
    m = R2Score()
    with pytest.raises(NotComputableError, match=r"R2Score must have at least one example before it can be computed"):
        m.compute()


def test_wrong_input_shapes():
    m = R2Score()

    with pytest.raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(4), torch.rand(4, 1)))

    with pytest.raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(4, 1), torch.rand(4)))


def test_r2_score(available_device):
    torch.manual_seed(42)
    size = 51

    y_pred = torch.rand(size)
    y = torch.rand(size)

    m = R2Score(device=available_device)
    assert m._device == torch.device(available_device)

    m.reset()
    m.update((y_pred, y))

    expected = r2_score(y.cpu().numpy(), y_pred.cpu().numpy())
    assert m.compute() == pytest.approx(expected)


def test_r2_score_2(available_device):
    torch.manual_seed(1)
    size = 105
    y_pred = torch.rand(size, 1)
    y = torch.rand(size, 1)

    y = y[torch.randperm(size)]

    m = R2Score(device=available_device)
    assert m._device == torch.device(available_device)

    m.reset()
    batch_size = 16
    n_iters = size // batch_size + 1
    for i in range(n_iters):
        idx = i * batch_size
        m.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))

    expected = r2_score(y.cpu().numpy(), y_pred.cpu().numpy())
    assert m.compute() == pytest.approx(expected)


def test_integration_r2_score(available_device):
    torch.manual_seed(1)
    size = 105
    y_pred = torch.rand(size, 1)
    y = torch.rand(size, 1)

    # Shuffle targets
    y = y[torch.randperm(size)]

    batch_size = 15

    def update_fn(engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        return y_pred[idx : idx + batch_size], y[idx : idx + batch_size]

    engine = Engine(update_fn)

    m = R2Score(device=available_device)
    assert m._device == torch.device(available_device)
    m.attach(engine, "r2_score")

    data = list(range(size // batch_size))
    r_squared = engine.run(data, max_epochs=1).metrics["r2_score"]

    expected = r2_score(y.cpu().numpy(), y_pred.cpu().numpy())
    assert r_squared == pytest.approx(expected)


def _test_distrib_compute(device, tol=1e-6):
    rank = idist.get_rank()

    def _test(metric_device):
        metric_device = torch.device(metric_device)
        m = R2Score(device=metric_device)

        y_pred = torch.randint(0, 10, size=(10,), device=device).float()
        y = torch.randint(0, 10, size=(10,), device=device).float()

        m.update((y_pred, y))

        # gather y_pred, y
        y_pred = idist.all_gather(y_pred)
        y = idist.all_gather(y)

        np_y_pred = y_pred.cpu().numpy()
        np_y = y.cpu().numpy()
        res = m.compute()
        assert r2_score(np_y, np_y_pred) == pytest.approx(res, abs=tol)

    for i in range(3):
        torch.manual_seed(10 + rank + i)
        _test("cpu")
        if device.type != "xla":
            _test(idist.device())


def _test_distrib_integration(device):
    rank = idist.get_rank()

    def _test(n_epochs, metric_device):
        metric_device = torch.device(metric_device)
        n_iters = 80
        batch_size = 16

        y_true = torch.randint(0, 10, size=(n_iters * batch_size,)).to(device).float()
        y_preds = torch.randint(0, 10, size=(n_iters * batch_size,)).to(device).float()

        def update(engine, i):
            return (
                y_preds[i * batch_size : (i + 1) * batch_size],
                y_true[i * batch_size : (i + 1) * batch_size],
            )

        engine = Engine(update)

        r2 = R2Score(device=metric_device)
        r2.attach(engine, "r2")

        data = list(range(n_iters))
        engine.run(data=data, max_epochs=n_epochs)

        y_preds = idist.all_gather(y_preds)
        y_true = idist.all_gather(y_true)

        assert "r2" in engine.state.metrics

        res = engine.state.metrics["r2"]
        if isinstance(res, torch.Tensor):
            res = res.cpu().numpy()

        true_res = r2_score(y_true.cpu().numpy(), y_preds.cpu().numpy())

        assert pytest.approx(res) == true_res

    metric_devices = ["cpu"]
    if device.type != "xla":
        metric_devices.append(idist.device())
    for metric_device in metric_devices:
        for i in range(2):
            torch.manual_seed(12 + rank + i)
            _test(n_epochs=1, metric_device=metric_device)
            _test(n_epochs=2, metric_device=metric_device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_compute, (device,), np=nproc, do_init=True)
    gloo_hvd_executor(_test_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_compute(device, tol=1e-3)
    _test_distrib_integration(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_compute(device, tol=1e-3)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



import os
import warnings
from collections import Counter

import pytest
import torch
from nltk.translate.bleu_score import corpus_bleu, sentence_bleu, SmoothingFunction

import ignite.distributed as idist
from ignite.exceptions import NotComputableError
from ignite.metrics.nlp import Bleu

from . import CorpusForTest

corpus = CorpusForTest(lower_split=True)


def test_wrong_inputs():
    with pytest.raises(ValueError, match=r"ngram order must be greater than zero"):
        Bleu(ngram=0)

    with pytest.raises(ValueError, match=r"Smooth is not valid"):
        Bleu(smooth="fake")

    with pytest.raises(ValueError, match=r"nb of candidates should be equal to nb of reference lists"):
        Bleu()._corpus_bleu(references=[[0], [0]], candidates=[[0]])

    with pytest.raises(NotComputableError):
        Bleu().compute()

    with pytest.raises(ValueError, match='Average must be either "macro" or "micro"'):
        Bleu(average="macros")


parametrize_args = (
    "candidates, references",
    [
        ([["a", "a", "a", "b", "c"]], [[["a", "b", "c"], ["a", "a", "d"]]]),
        corpus.sample_1,
        corpus.sample_2,
        corpus.sample_3,
        corpus.sample_4,
    ],
)


def _test(candidates, references, average, smooth="no_smooth", smooth_nltk_fn=None, ngram_range=8, device="cpu"):
    for i in range(1, ngram_range):
        weights = tuple([1 / i] * i)
        bleu = Bleu(ngram=i, average=average, smooth=smooth, device=device)

        if average == "macro":
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                reference = sentence_bleu(
                    references[0], candidates[0], weights=weights, smoothing_function=smooth_nltk_fn
                )
            assert pytest.approx(reference) == bleu._sentence_bleu(references[0], candidates[0])

        elif average == "micro":
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                reference = corpus_bleu(references, candidates, weights=weights, smoothing_function=smooth_nltk_fn)
            assert pytest.approx(reference) == bleu._corpus_bleu(references, candidates)

        bleu.update((candidates, references))
        computed = bleu.compute()
        if isinstance(computed, torch.Tensor):
            computed = computed.cpu().item()
        assert pytest.approx(reference) == computed


@pytest.mark.parametrize(*parametrize_args)
def test_macro_bleu(candidates, references, available_device):
    _test(candidates, references, "macro", device=available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_micro_bleu(candidates, references, available_device):
    _test(candidates, references, "micro", device=available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_macro_bleu_smooth1(candidates, references, available_device):
    _test(candidates, references, "macro", "smooth1", SmoothingFunction().method1, device=available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_micro_bleu_smooth1(candidates, references, available_device):
    _test(candidates, references, "micro", "smooth1", SmoothingFunction().method1, device=available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_macro_bleu_nltk_smooth2(candidates, references, available_device):
    _test(candidates, references, "macro", "nltk_smooth2", SmoothingFunction().method2, device=available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_micro_bleu_nltk_smooth2(candidates, references, available_device):
    _test(candidates, references, "micro", "nltk_smooth2", SmoothingFunction().method2, device=available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_macro_bleu_smooth2(candidates, references, available_device):
    _test(candidates, references, "macro", "smooth2", SmoothingFunction().method2, 3, available_device)


@pytest.mark.parametrize(*parametrize_args)
def test_micro_bleu_smooth2(candidates, references, available_device):
    _test(candidates, references, "micro", "smooth2", SmoothingFunction().method2, 3, device=available_device)


def test_accumulation_macro_bleu(available_device):
    bleu = Bleu(ngram=4, smooth="smooth2", device=available_device)
    assert bleu._device == torch.device(available_device)

    bleu.update(([corpus.cand_1], [corpus.references_1]))
    bleu.update(([corpus.cand_2a], [corpus.references_2]))
    bleu.update(([corpus.cand_2b], [corpus.references_2]))
    bleu.update(([corpus.cand_3], [corpus.references_2]))
    value = bleu._sentence_bleu(corpus.references_1, corpus.cand_1)
    value += bleu._sentence_bleu(corpus.references_2, corpus.cand_2a)
    value += bleu._sentence_bleu(corpus.references_2, corpus.cand_2b)
    value += bleu._sentence_bleu(corpus.references_2, corpus.cand_3)
    assert bleu.compute() == value / 4


def test_accumulation_micro_bleu(available_device):
    bleu = Bleu(ngram=4, smooth="smooth2", average="micro", device=available_device)
    assert bleu._device == torch.device(available_device)

    bleu.update(([corpus.cand_1], [corpus.references_1]))
    bleu.update(([corpus.cand_2a], [corpus.references_2]))
    bleu.update(([corpus.cand_2b], [corpus.references_2]))
    bleu.update(([corpus.cand_3], [corpus.references_2]))
    value = bleu._corpus_bleu(
        [corpus.references_1, corpus.references_2, corpus.references_2, corpus.references_2],
        [corpus.cand_1, corpus.cand_2a, corpus.cand_2b, corpus.cand_3],
    )
    assert bleu.compute() == value


def test_bleu_batch_macro(available_device):
    bleu = Bleu(ngram=4, device=available_device)
    assert bleu._device == torch.device(available_device)

    # Batch size 3
    hypotheses = [corpus.cand_1, corpus.cand_2a, corpus.cand_2b]
    refs = [corpus.references_1, corpus.references_2, corpus.references_2]
    bleu.update((hypotheses, refs))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        reference_bleu_score = (
            sentence_bleu(refs[0], hypotheses[0])
            + sentence_bleu(refs[1], hypotheses[1])
            + sentence_bleu(refs[2], hypotheses[2])
        ) / 3
    computed = bleu.compute()
    if isinstance(computed, torch.Tensor):
        computed = computed.cpu().item()

    assert pytest.approx(computed) == reference_bleu_score

    value = 0
    for _hypotheses, _refs in zip(hypotheses, refs):
        value += bleu._sentence_bleu(_refs, _hypotheses)
        bleu.update(([_hypotheses], [_refs]))

    ref_1 = value / len(refs)
    computed = bleu.compute()
    if isinstance(computed, torch.Tensor):
        computed = computed.cpu().item()

    assert pytest.approx(ref_1) == reference_bleu_score
    assert pytest.approx(computed) == reference_bleu_score


def test_bleu_batch_micro(available_device):
    bleu = Bleu(ngram=4, average="micro", device=available_device)
    assert bleu._device == torch.device(available_device)

    # Batch size 3
    hypotheses = [corpus.cand_1, corpus.cand_2a, corpus.cand_2b]
    refs = [corpus.references_1, corpus.references_2, corpus.references_2]
    bleu.update((hypotheses, refs))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        reference_bleu_score = corpus_bleu(refs, hypotheses)
    assert pytest.approx(bleu.compute()) == reference_bleu_score

    assert pytest.approx(bleu._corpus_bleu(refs, hypotheses)) == reference_bleu_score


@pytest.mark.parametrize(
    "candidates, references",
    [
        (corpus.cand_1, corpus.references_1),
        (corpus.cand_2a, corpus.references_2),
        (corpus.cand_2b, corpus.references_2),
        (corpus.cand_1, corpus.references_1),
    ],
)
def test_n_gram_counter(candidates, references, available_device):
    bleu = Bleu(ngram=4, device=available_device)
    assert bleu._device == torch.device(available_device)

    hyp_length, ref_length = bleu._n_gram_counter([references], [candidates], Counter(), Counter())
    assert hyp_length == len(candidates)

    ref_lens = (len(reference) for reference in references)
    closest_ref_len = min(ref_lens, key=lambda ref_len: (abs(ref_len - len(candidates)), ref_len))

    assert ref_length == closest_ref_len


def _test_macro_distrib_integration(device):
    from ignite.engine import Engine

    rank = idist.get_rank()

    size = len(corpus.chunks)

    data = []
    for c in corpus.chunks:
        data += idist.get_world_size() * [c]

    def update(_, i):
        return data[i + size * rank]

    def _test(device):
        engine = Engine(update)
        m = Bleu(ngram=4, smooth="smooth2", device=device)
        m.attach(engine, "bleu")

        engine.run(data=list(range(size)), max_epochs=1)

        assert "bleu" in engine.state.metrics

        ref_bleu = 0
        for candidates, references in data:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                ref_bleu += sentence_bleu(
                    references[0],
                    candidates[0],
                    weights=[0.25, 0.25, 0.25, 0.25],
                    smoothing_function=SmoothingFunction().method2,
                )

        assert pytest.approx(engine.state.metrics["bleu"]) == ref_bleu / len(data)

    _test("cpu")

    if device.type != "xla":
        _test(idist.device())


def _test_micro_distrib_integration(device):
    from ignite.engine import Engine

    rank = idist.get_rank()

    size = len(corpus.chunks)

    data = []
    for c in corpus.chunks:
        data += idist.get_world_size() * [c]

    def update(_, i):
        return data[i + size * rank]

    def _test(metric_device):
        engine = Engine(update)
        m = Bleu(ngram=4, smooth="smooth2", average="micro", device=metric_device)
        m.attach(engine, "bleu")

        engine.run(data=list(range(size)), max_epochs=1)

        assert "bleu" in engine.state.metrics

        ref_bleu = 0
        references = []
        candidates = []
        for _candidates, _references in data:
            references.append(_references[0])
            candidates.append(_candidates[0])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ref_bleu += corpus_bleu(
                references,
                candidates,
                weights=[0.25, 0.25, 0.25, 0.25],
                smoothing_function=SmoothingFunction().method2,
            )

        assert pytest.approx(engine.state.metrics["bleu"]) == ref_bleu

    _test("cpu")

    if device.type != "xla":
        _test(idist.device())


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_macro_distrib_integration(device)
    _test_micro_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_macro_distrib_integration(device)
    _test_micro_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_macro_distrib_integration, (device,), np=nproc, do_init=True)
    gloo_hvd_executor(_test_micro_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_macro_distrib_integration(device)
    _test_micro_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_macro_distrib_integration(device)
    _test_micro_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_macro_distrib_integration(device)
    _test_micro_distrib_integration(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_macro_distrib_integration(device)
    _test_micro_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



from unittest.mock import MagicMock

from ignite.engine import Engine, Events
from ignite.handlers import global_step_from_engine


def test_global_step_from_engine():
    iteration = 12
    epoch = 23

    trainer = Engine(lambda e, b: None)
    trainer.state.iteration = iteration
    trainer.state.epoch = epoch

    gst = global_step_from_engine(trainer)
    assert gst(MagicMock(), Events.EPOCH_COMPLETED) == epoch

    gst = global_step_from_engine(trainer, custom_event_name=Events.ITERATION_COMPLETED)
    assert gst(MagicMock(), Events.EPOCH_COMPLETED) == iteration



import numpy as np
import pytest
import torch
from sklearn.metrics import average_precision_score, precision_recall_curve

from ignite import distributed as idist
from ignite.engine import Engine
from ignite.metrics import MeanAveragePrecision
from ignite.utils import to_onehot


def test_wrong_input():
    with pytest.raises(ValueError, match="rec_thresholds should be a one-dimensional tensor or a sequence of floats"):
        MeanAveragePrecision(rec_thresholds=torch.zeros((2, 2)))

    with pytest.raises(TypeError, match="rec_thresholds should be a sequence of floats or a tensor"):
        MeanAveragePrecision(rec_thresholds={0, 0.2, 0.4, 0.6, 0.8})

    with pytest.raises(ValueError, match="Wrong `class_mean` parameter"):
        MeanAveragePrecision(class_mean="samples")

    with pytest.raises(ValueError, match="rec_thresholds values should be between 0 and 1"):
        MeanAveragePrecision(rec_thresholds=(0.0, 0.5, 1.0, 1.5))

    metric = MeanAveragePrecision()
    with pytest.raises(RuntimeError, match="Metric could not be computed without any update method call"):
        metric.compute()


def test_wrong_classification_input():
    metric = MeanAveragePrecision()

    with pytest.raises(TypeError, match="`y_pred` should be a float tensor"):
        metric.update((torch.tensor([0, 1, 0]), torch.tensor([1, 0, 1])))

    metric = MeanAveragePrecision()
    with pytest.warns(RuntimeWarning, match="`y` should be of dtype long when entry type is multiclass"):
        metric.update((torch.tensor([[0.5, 0.4, 0.1]]), torch.tensor([2.0])))

    with pytest.raises(ValueError, match="y_pred contains fewer classes than y"):
        metric.update((torch.tensor([[0.5, 0.4, 0.1]]), torch.tensor([3])))


def test__prepare_output():
    metric = MeanAveragePrecision()

    metric._type = "binary"
    scores, y = metric._prepare_output((torch.rand((5, 4, 3, 2)), torch.randint(0, 2, (5, 4, 3, 2)).bool()))
    assert scores.shape == y.shape == (1, 120)

    metric._type = "multiclass"
    scores, y = metric._prepare_output((torch.rand((5, 4, 3, 2)), torch.randint(0, 4, (5, 3, 2))))
    assert scores.shape == (4, 30) and y.shape == (30,)

    metric._type = "multilabel"
    scores, y = metric._prepare_output((torch.rand((5, 4, 3, 2)), torch.randint(0, 2, (5, 4, 3, 2)).bool()))
    assert scores.shape == y.shape == (4, 30)


def test_update():
    metric = MeanAveragePrecision()
    assert len(metric._y_pred) == len(metric._y_true) == 0
    metric.update((torch.rand((5, 4)), torch.randint(0, 2, (5, 4)).bool()))
    assert len(metric._y_pred) == len(metric._y_true) == 1


def test__compute_recall_and_precision():
    m = MeanAveragePrecision()

    scores = torch.rand((50,))
    y_true = torch.randint(0, 2, (50,)).bool()
    precision, recall, _ = precision_recall_curve(y_true.numpy(), scores.numpy())
    P = y_true.sum(dim=-1)
    ignite_recall, ignite_precision = m._compute_recall_and_precision(y_true, scores, P)
    assert (ignite_recall.squeeze().flip(0).numpy() == recall[:-1]).all()
    assert (ignite_precision.squeeze().flip(0).numpy() == precision[:-1]).all()

    # When there's no actual positive. Numpy expectedly raises warning.
    scores = torch.rand((50,))
    y_true = torch.zeros((50,)).bool()
    precision, recall, _ = precision_recall_curve(y_true.numpy(), scores.numpy())
    P = torch.tensor(0)
    ignite_recall, ignite_precision = m._compute_recall_and_precision(y_true, scores, P)
    assert (ignite_recall.flip(0).numpy() == recall[:-1]).all()
    assert (ignite_precision.flip(0).numpy() == precision[:-1]).all()


def test__compute_average_precision():
    m = MeanAveragePrecision()

    # Binary data
    scores = np.random.rand(50)
    y_true = np.random.randint(0, 2, 50)
    ap = average_precision_score(y_true, scores)
    precision, recall, _ = precision_recall_curve(y_true, scores)
    ignite_ap = m._compute_average_precision(
        torch.from_numpy(recall[:-1]).flip(-1), torch.from_numpy(precision[:-1]).flip(-1)
    )
    assert np.allclose(ignite_ap.item(), ap)

    # Multilabel data
    scores = np.random.rand(50, 5)
    y_true = np.random.randint(0, 2, (50, 5))
    ap = average_precision_score(y_true, scores, average=None)
    ignite_ap = []
    for cls in range(scores.shape[1]):
        precision, recall, _ = precision_recall_curve(y_true[:, cls], scores[:, cls])
        ignite_ap.append(
            m._compute_average_precision(
                torch.from_numpy(recall[:-1]).flip(-1), torch.from_numpy(precision[:-1]).flip(-1)
            ).item()
        )
    ignite_ap = np.array(ignite_ap)
    assert np.allclose(ignite_ap, ap)


def test_compute_binary_data():
    m = MeanAveragePrecision()
    scores = torch.rand((130,))
    y_true = torch.randint(0, 2, (130,))

    m.update((scores[:50], y_true[:50]))
    m.update((scores[50:], y_true[50:]))
    ignite_map = m.compute()

    map = average_precision_score(y_true.numpy(), scores.numpy())

    assert np.allclose(ignite_map, map)


@pytest.mark.parametrize("class_mean", [None, "macro", "micro", "weighted"])
def test_compute_nonbinary_data(class_mean):
    scores = torch.rand((130, 5, 2, 2))
    sklearn_scores = scores.transpose(1, -1).reshape(-1, 5).numpy()

    # Multiclass
    m = MeanAveragePrecision(class_mean=class_mean)
    y_true = torch.randint(0, 5, (130, 2, 2))
    m.update((scores[:50], y_true[:50]))
    m.update((scores[50:], y_true[50:]))
    ignite_map = m.compute().numpy()

    y_true = to_onehot(y_true, 5).transpose(1, -1).reshape(-1, 5).numpy()
    sklearn_map = average_precision_score(y_true, sklearn_scores, average=class_mean)

    assert np.allclose(sklearn_map, ignite_map)

    # Multilabel
    m = MeanAveragePrecision(is_multilabel=True, class_mean=class_mean)
    y_true = torch.randint(0, 2, (130, 5, 2, 2)).bool()
    m.update((scores[:50], y_true[:50]))
    m.update((scores[50:], y_true[50:]))
    ignite_map = m.compute().numpy()

    y_true = y_true.transpose(1, -1).reshape(-1, 5).numpy()
    sklearn_map = average_precision_score(y_true, sklearn_scores, average=class_mean)

    assert np.allclose(sklearn_map, ignite_map)


@pytest.mark.parametrize("data_type", ["binary", "multiclass", "multilabel"])
@pytest.mark.parametrize("n_epochs", [1, 2])
def test_distrib_integration(distributed, data_type, n_epochs):
    rank = idist.get_rank()
    device = idist.device()
    torch.manual_seed(12 + rank)

    n_iters = 60
    batch_size = 16
    n_classes = 7

    metric_devices = [torch.device("cpu")]
    if device.type != "xla":
        metric_devices.append(device)

    for metric_device in metric_devices:

        y_true_size = (
            (n_iters * batch_size, 3, 2) if data_type != "multilabel" else (n_iters * batch_size, n_classes, 3, 2)
        )
        y_true = torch.randint(0, n_classes if data_type == "multiclass" else 2, size=y_true_size, device=device)
        y_preds_size = (
            (n_iters * batch_size, n_classes, 3, 2) if data_type != "binary" else (n_iters * batch_size, 3, 2)
        )
        y_preds = torch.rand(y_preds_size, device=device)

        def update(_, i):
            return (
                y_preds[i * batch_size : (i + 1) * batch_size, ...],
                y_true[i * batch_size : (i + 1) * batch_size, ...],
            )

        engine = Engine(update)
        mAP = MeanAveragePrecision(is_multilabel=data_type == "multilabel", device=metric_device)
        mAP.attach(engine, "mAP")

        engine.run(range(n_iters), max_epochs=n_epochs)

        y_preds = idist.all_gather(y_preds)
        y_true = idist.all_gather(y_true)

        assert "mAP" in engine.state.metrics

        if data_type == "multiclass":
            y_true = to_onehot(y_true, n_classes)

        if data_type == "binary":
            y_true = y_true.view(-1)
            y_preds = y_preds.view(-1)
        else:
            y_true = y_true.transpose(1, -1).reshape(-1, n_classes)
            y_preds = y_preds.transpose(1, -1).reshape(-1, n_classes)

        sklearn_mAP = average_precision_score(y_true.cpu().numpy(), y_preds.cpu().numpy())
        assert np.allclose(sklearn_mAP, engine.state.metrics["mAP"])



from typing import Tuple

import numpy as np
import pytest

import torch
from sklearn.metrics import calinski_harabasz_score
from torch import Tensor

from ignite import distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.clustering import CalinskiHarabaszScore


def test_zero_sample():
    with pytest.raises(
        NotComputableError, match="CalinskiHarabaszScore must have at least one example before it can be computed"
    ):
        metric = CalinskiHarabaszScore()
        metric.compute()


def test_wrong_output_shape():
    wrong_features = torch.zeros(4, dtype=torch.float)
    correct_features = torch.zeros(4, 3, dtype=torch.float)
    wrong_labels = torch.tensor([[0, 0, 1, 1], [0, 0, 1, 1]], dtype=torch.long)
    correct_labels = torch.tensor([0, 0, 1, 1], dtype=torch.long)

    with pytest.raises(ValueError, match=r"Features should be of shape \(batch_size, n_targets\)"):
        metric = CalinskiHarabaszScore()
        metric.update((wrong_features, correct_labels))

    with pytest.raises(ValueError, match=r"Labels should be of shape \(batch_size, \)"):
        metric = CalinskiHarabaszScore()
        metric.update((correct_features, wrong_labels))


def test_wrong_output_dtype():
    wrong_features = torch.zeros(4, 3, dtype=torch.long)
    correct_features = torch.zeros(4, 3, dtype=torch.float)
    wrong_labels = torch.tensor([0, 0, 1, 1], dtype=torch.float)
    correct_labels = torch.tensor([0, 0, 1, 1], dtype=torch.long)

    with pytest.raises(ValueError, match=r"Incoherent types between input features and stored features"):
        metric = CalinskiHarabaszScore()
        metric.update((correct_features, correct_labels))
        metric.update((wrong_features, correct_labels))

    with pytest.raises(ValueError, match=r"Incoherent types between input labels and stored labels"):
        metric = CalinskiHarabaszScore()
        metric.update((correct_features, correct_labels))
        metric.update((correct_features, wrong_labels))


@pytest.fixture(params=list(range(2)))
def test_case(request):
    N = 100
    NDIM = 10
    BS = 10

    # well-clustered case
    random_order = torch.from_numpy(np.random.permutation(N * 3))
    x1 = torch.cat(
        [
            torch.normal(-5.0, 1.0, size=(N, NDIM)),
            torch.normal(5.0, 1.0, size=(N, NDIM)),
            torch.normal(0.0, 1.0, size=(N, NDIM)),
        ]
    ).float()[random_order]
    y1 = torch.tensor([0] * N + [1] * N + [2] * N, dtype=torch.long)[random_order]

    # poorly-clustered case
    x2 = torch.cat(
        [
            torch.normal(-1.0, 1.0, size=(N, NDIM)),
            torch.normal(0.0, 1.0, size=(N, NDIM)),
            torch.normal(1.0, 1.0, size=(N, NDIM)),
        ]
    ).float()
    y2 = torch.from_numpy(np.random.choice(3, size=N * 3)).long()

    return [
        (x1, y1, BS),
        (x2, y2, BS),
    ][request.param]


@pytest.mark.parametrize("n_times", range(5))
def test_integration(n_times: int, test_case: Tuple[Tensor, Tensor, Tensor], available_device):
    features, labels, batch_size = test_case

    np_features = features.numpy()
    np_labels = labels.numpy()

    def update_fn(engine: Engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        feature_batch = np_features[idx : idx + batch_size]
        label_batch = np_labels[idx : idx + batch_size]
        return torch.from_numpy(feature_batch), torch.from_numpy(label_batch)

    engine = Engine(update_fn)

    m = CalinskiHarabaszScore(device=available_device)
    assert m._device == torch.device(available_device)

    m.attach(engine, "calinski_harabasz_score")

    data = list(range(np_features.shape[0] // batch_size))
    s = engine.run(data, max_epochs=1).metrics["calinski_harabasz_score"]

    np_ans = calinski_harabasz_score(np_features, np_labels)

    assert pytest.approx(np_ans, rel=1e-5) == s


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_compute(self):
        rank = idist.get_rank()
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        torch.manual_seed(10 + rank)
        for metric_device in metric_devices:
            m = CalinskiHarabaszScore(device=metric_device)

            random_order = torch.from_numpy(np.random.permutation(200))
            features = torch.cat([torch.normal(-1.0, 1.0, size=(100, 10)), torch.normal(1.0, 1.0, size=(100, 10))]).to(
                device
            )[random_order]
            labels = torch.tensor([0] * 100 + [1] * 100, dtype=torch.long, device=device)[random_order]

            m.update((features, labels))

            features = idist.all_gather(features)
            labels = idist.all_gather(labels)

            np_features = features.cpu().numpy()
            np_labels = labels.cpu().numpy()

            np_ans = calinski_harabasz_score(np_features, np_labels)

            assert pytest.approx(np_ans, rel=1e-5) == m.compute()

    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration(self, n_epochs: int):
        tol = 1e-5
        rank = idist.get_rank()
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        n_iters = 80
        batch_size = 16

        for metric_device in metric_devices:
            torch.manual_seed(12 + rank)

            cluster_size = n_iters * batch_size // 2
            random_order = torch.from_numpy(np.random.permutation(n_iters * batch_size))
            features = torch.cat(
                [torch.normal(-1.0, 1.0, size=(cluster_size, 10)), torch.normal(1.0, 1.0, size=(cluster_size, 10))]
            ).to(device)[random_order]
            labels = torch.tensor([0] * cluster_size + [1] * cluster_size, dtype=torch.long, device=device)[
                random_order
            ]

            engine = Engine(
                lambda e, i: (
                    features[i * batch_size : (i + 1) * batch_size],
                    labels[i * batch_size : (i + 1) * batch_size],
                )
            )

            chs = CalinskiHarabaszScore(device=metric_device)
            chs.attach(engine, "calinski_harabasz_score")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            features = idist.all_gather(features)
            labels = idist.all_gather(labels)

            assert "calinski_harabasz_score" in engine.state.metrics

            res = engine.state.metrics["calinski_harabasz_score"]

            np_labels = labels.cpu().numpy()
            np_features = features.cpu().numpy()

            np_ans = calinski_harabasz_score(np_features, np_labels)

            assert pytest.approx(np_ans, rel=tol) == res



import torch.nn as nn
from transformers import AutoConfig, AutoModelForSequenceClassification


class TransformerModel(nn.Module):
    def __init__(self, model_name, model_dir, dropout, n_fc, n_classes):
        super(TransformerModel, self).__init__()
        self.config = AutoConfig.from_pretrained(
            model_name,
            num_labels=n_classes,
            output_hidden_states=n_fc,
            classifier_dropout=dropout,
            output_attentions=True,
        )
        self.transformer = AutoModelForSequenceClassification.from_pretrained(
            model_name, cache_dir=model_dir, config=self.config
        )

    def forward(self, inputs):
        output = self.transformer(**inputs)["logits"]

        return output



import dill

from ignite.metrics import Metric


class Accumulation(Metric):
    def __init__(self):
        self.value = 0
        super(Accumulation, self).__init__()

    def reset(self):
        self.value = 0

    def compute(self):
        return self.value

    def update(self, output):
        self.value += output


def test_metric():
    def _test(m, values, e):
        for v in values:
            m.update(v)
        assert m.compute() == e

    metric = Accumulation()

    m1 = dill.loads(dill.dumps(metric))

    values = list(range(10))
    expected = sum(values)

    _test(m1, values, expected)

    metric.update(5)

    m2 = dill.loads(dill.dumps(metric))

    _test(m2, values, expected + 5)



import torch
from dataset import TransformerDataset
from datasets import load_dataset
from model import TransformerModel
from transformers import AutoTokenizer

from ignite.handlers import DiskSaver


def get_tokenizer(tokenizer_name, tokenizer_dir):
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_name, cache_dir=tokenizer_dir, do_lower_case=True)
    return tokenizer


def get_model(model_name, model_dir, drop_out, n_fc, num_classes):
    model = TransformerModel(model_name, model_dir, drop_out, n_fc, num_classes)
    return model


def get_dataset(cache_dir, tokenizer_name, tokenizer_dir, max_length):
    train_dataset, test_dataset = load_dataset("imdb", split=["train", "test"], cache_dir=cache_dir)
    tokenizer = get_tokenizer(tokenizer_name, tokenizer_dir)
    train_texts, train_labels = train_dataset["text"], train_dataset["label"]
    test_texts, test_labels = test_dataset["text"], test_dataset["label"]
    train_dataset = TransformerDataset(train_texts, train_labels, tokenizer, max_length)
    test_dataset = TransformerDataset(test_texts, test_labels, tokenizer, max_length)
    return train_dataset, test_dataset


def thresholded_output_transform(output):
    y_pred, y = output
    return torch.round(torch.sigmoid(y_pred)), y


def get_save_handler(config):
    if config["with_clearml"]:
        from ignite.handlers.clearml_logger import ClearMLSaver

        return ClearMLSaver(dirname=config["output_dir"])

    return DiskSaver(config["output_dir"], require_empty=False)



import os
from typing import Any, Callable, Union

import pytest
import torch
import torch.nn as nn
from torch.nn.parallel import DataParallel, DistributedDataParallel

import ignite.distributed as idist
from ignite.engine import Engine, Events
from ignite.handlers import EMAHandler


def _get_dummy_model() -> nn.Module:
    model = nn.Linear(2, 1, bias=False)
    model.weight.data.fill_(1)
    return model


def _unwrap_model(model):
    if isinstance(model, (DataParallel, DistributedDataParallel)):
        return model.module
    else:
        return model


@pytest.fixture(scope="module")
def get_dummy_model():
    """Returns a function since the fixture is needed multiple times in a single test"""
    yield _get_dummy_model


def _get_dummy_step_fn(model: Union[nn.Module, DataParallel, DistributedDataParallel]) -> Callable:
    """Get a dummy step function, given model is a (wrapper of) dummy model returned from _get_dummy_model"""

    def step_fn(engine, batch):
        """Increment the weight by 1 at each iteration"""
        _unwrap_model(model).weight.data.add_(1)
        return 0

    return step_fn


@pytest.mark.parametrize("momentum", [-1, 2])
def test_ema_invalid_momentum(get_dummy_model, momentum):
    with pytest.raises(ValueError, match="Invalid momentum"):
        EMAHandler(get_dummy_model(), momentum=momentum)


def test_has_momentum_scheduler(get_dummy_model):
    """Test the handler has attribute `momentum_scheduler` and `_momentum_lambda_obj`"""
    momentum_warmup = 0.0
    warmup_iters = 10
    ema_handler = EMAHandler(get_dummy_model(), momentum_warmup=momentum_warmup, warmup_iters=warmup_iters)
    assert hasattr(ema_handler, "momentum_scheduler")
    assert hasattr(ema_handler, "_momentum_lambda_obj")


def test_ema_warmup_func(get_dummy_model):
    """Test the built-in linear warmup function for the EMA momentum"""
    momentum = 0.5
    momentum_warmup_1 = 0.0
    momentum_warmup_2 = 1.0
    warmup_iters = 5

    def check_ema_momentum(engine: Engine, momentum_warmup, final_momentum, warmup_iters):
        if engine.state.iteration == 1:
            assert engine.state.ema_momentum == momentum_warmup
        elif engine.state.iteration >= 1 + warmup_iters:
            assert engine.state.ema_momentum == final_momentum
        else:
            min_momentum = min(momentum, momentum_warmup)
            max_momentum = max(momentum, momentum_warmup)
            assert min_momentum <= engine.state.ema_momentum <= max_momentum

    # momentum_warmup < momentum
    model_1 = get_dummy_model()
    engine_1 = Engine(_get_dummy_step_fn(model_1))
    ema_handler_1 = EMAHandler(model_1, momentum, momentum_warmup_1, warmup_iters)
    ema_handler_1.attach(engine_1)
    engine_1.add_event_handler(
        Events.ITERATION_COMPLETED, check_ema_momentum, momentum_warmup_1, momentum, warmup_iters
    )
    engine_1.run(range(10))

    # momentum_warmup > momentum
    model_2 = get_dummy_model()
    engine_2 = Engine(_get_dummy_step_fn(model_2))
    ema_handler_2 = EMAHandler(model_2, momentum, momentum_warmup_2, warmup_iters)
    ema_handler_2.attach(engine_2)
    engine_2.add_event_handler(
        Events.ITERATION_COMPLETED, check_ema_momentum, momentum_warmup_2, momentum, warmup_iters
    )
    engine_2.run(range(10))


def test_ema_invalid_model():
    with pytest.raises(ValueError, match="model should be an instance of nn.Module or its subclasses"):
        model = "Invalid Model"
        EMAHandler(model)  # type: ignore


@pytest.mark.distributed
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_ema_ema_model_on_cuda(get_dummy_model):
    """Test if ema_handler.ema_model is nn.Module or nn.DataParallel and under eval mode"""
    model = get_dummy_model().to(idist.device())
    model = idist.auto_model(model)
    ema_handler = EMAHandler(model)
    ema_model = ema_handler.ema_model
    assert not ema_model.training
    if isinstance(model, DataParallel):
        assert isinstance(ema_model, DataParallel)
    else:
        assert (
            isinstance(ema_model, nn.Module)
            and (not isinstance(ema_model, DataParallel))
            and (not isinstance(ema_model, DistributedDataParallel))
        )


def test_ema_load_state_dict(get_dummy_model):
    model_1 = get_dummy_model()
    model_1.weight.data.fill_(2)
    state_dict_1 = model_1.state_dict()

    model_2 = get_dummy_model()
    ema_handler = EMAHandler(model_2)
    ema_model = ema_handler.ema_model
    ema_model.load_state_dict(state_dict_1)
    assert ema_model.weight.data.allclose(model_1.weight.data)


def test_ema_get_const_momentum(get_dummy_model):
    """Test if momentum retrieved from the engine is constant and equal to the handler's momentum"""
    model = get_dummy_model()
    step_fn = _get_dummy_step_fn(model)
    engine = Engine(step_fn)

    def assert_const_momentum(engine: Engine, const_momentum):
        assert engine.state.ema_momentum == const_momentum

    ema_handler = EMAHandler(model, momentum=0.002)
    ema_handler.attach(engine)
    engine.add_event_handler(Events.ITERATION_COMPLETED, assert_const_momentum, ema_handler.momentum)
    engine.run(range(10))


@pytest.mark.parametrize("handle_buffers", ["copy", "update", "ema_train", "invalid"])
def test_ema_buffer(handle_buffers):
    """Test if the tensors in buffer are also correctly updated"""
    model = nn.BatchNorm2d(2)
    model.running_mean.data.fill_(1.5)
    model.running_var.data.fill_(1.5)

    # manually register a buffer to test if it will be correctly updated
    model.register_buffer("dummy_buffer", tensor=torch.tensor(1.0, dtype=torch.float32))

    if handle_buffers == "invalid":
        with pytest.raises(ValueError, match="handle_buffers can only"):
            _ = EMAHandler(model, momentum=0.5, handle_buffers=handle_buffers)
    else:
        ema_handler = EMAHandler(model, momentum=0.5, handle_buffers=handle_buffers)

        def _bn_step_fn(engine, batch):
            x = torch.rand(4, 2, 32, 32)
            _ = model(x)
            # manually increment the dummy_buffer at every step
            model.dummy_buffer += 1.0
            return 1

        engine = Engine(_bn_step_fn)
        ema_handler.attach(engine)

        ema_model = ema_handler.ema_model
        if handle_buffers == "ema_train":
            assert ema_model.training
        else:
            assert not ema_model.training

        @engine.on(Events.ITERATION_COMPLETED)
        def check_buffers():
            if handle_buffers == "update":
                # the buffers with torch.int64 data type should be directly copied
                assert ema_model.num_batches_tracked.allclose(model.num_batches_tracked)

                # buffers with floating type will be updated rather than copied
                assert not ema_model.dummy_buffer.allclose(model.dummy_buffer)
                assert not ema_model.running_mean.allclose(model.running_mean)
                assert not ema_model.running_var.allclose(model.running_var)
            elif handle_buffers == "copy":
                # the buffers with torch.int64 data type should be directly copied
                assert ema_model.num_batches_tracked.allclose(model.num_batches_tracked)

                assert ema_model.dummy_buffer.allclose(model.dummy_buffer)
                assert ema_model.running_mean.allclose(model.running_mean)
                assert ema_model.running_var.allclose(model.running_var)
            else:
                # buffers will not be copied or EMA updated
                assert ema_model.num_batches_tracked.allclose(torch.tensor(0, dtype=torch.int64))
                assert ema_model.dummy_buffer.allclose(torch.tensor(1.0, dtype=torch.float32))

        # engine will run 4 iterations
        engine.run([0, 1], max_epochs=2)

        if handle_buffers == "update":
            assert ema_model.num_batches_tracked.allclose(model.num_batches_tracked)
            assert ema_model.dummy_buffer.allclose(torch.tensor(4.0625, dtype=torch.float32))
            assert not ema_model.dummy_buffer.allclose(model.dummy_buffer)
            assert not ema_model.running_mean.allclose(model.running_mean)
            assert not ema_model.running_var.allclose(model.running_var)
        elif handle_buffers == "copy":
            assert ema_model.num_batches_tracked.allclose(model.num_batches_tracked)
            assert ema_model.dummy_buffer.allclose(model.dummy_buffer)
            assert ema_model.running_mean.allclose(model.running_mean)
            assert ema_model.running_var.allclose(model.running_var)
        else:
            # buffers will not be copied or EMA updated
            assert ema_model.num_batches_tracked.allclose(torch.tensor(0, dtype=torch.int64))
            assert ema_model.dummy_buffer.allclose(torch.tensor(1.0, dtype=torch.float32))


def test_ema_two_handlers(get_dummy_model):
    """Test when two EMA handlers are attached to a trainer"""
    model_1 = get_dummy_model()
    ema_handler_1 = EMAHandler(model_1, momentum=0.5)

    model_2 = get_dummy_model()
    ema_handler_2 = EMAHandler(model_2, momentum=0.5)

    def _step_fn(engine: Engine, batch: Any):
        model_1.weight.data.add_(1)
        model_2.weight.data.add_(1)
        return 0

    engine = Engine(_step_fn)
    assert not hasattr(engine.state, "ema_momentum_1")
    # handler_1 update EMA model of model_1 every 1 iteration
    ema_handler_1.attach(engine, "ema_momentum_1", event=Events.ITERATION_COMPLETED)
    assert hasattr(engine.state, "ema_momentum_1")

    # handler_2 update EMA model for model_2 every 2 iterations
    ema_handler_2.attach(engine, "ema_momentum_2", event=Events.ITERATION_COMPLETED(every=2))
    assert hasattr(engine.state, "ema_momentum_2")

    # engine will run 4 iterations
    engine.run(range(2), max_epochs=2)
    # explicitly cast to float32 to avoid test failure on XLA devices
    ema_weight_1 = ema_handler_1.ema_model.weight.data.to(torch.float32)
    ema_weight_2 = ema_handler_2.ema_model.weight.data.to(torch.float32)
    assert ema_weight_1.allclose(ema_weight_1.new_full((1, 2), 4.0625))
    assert ema_weight_2.allclose(ema_weight_2.new_full((1, 2), 3.5))

    assert engine.state.ema_momentum_1 == 0.5
    assert engine.state.ema_momentum_2 == 0.5

    model_3 = get_dummy_model()
    ema_handler_3 = EMAHandler(model_3)
    with pytest.warns(UserWarning, match="Attribute 'ema_momentum_1' already exists"):
        ema_handler_3.attach(engine, name="ema_momentum_1")


def _test_ema_final_weight(model, device=None, ddp=False, interval=1):
    """Test if final smoothed weights are correct"""
    if device is None:
        # let horovod decide the device
        device = idist.device()
    if isinstance(device, str):
        device = torch.device(device)
    model = model.to(device)
    if ddp:
        model = idist.auto_model(model)
    step_fn = _get_dummy_step_fn(model)
    engine = Engine(step_fn)

    ema_handler = EMAHandler(model, momentum=0.5)
    ema_handler.attach(engine, "model", event=Events.ITERATION_COMPLETED(every=interval))

    # engine will run 4 iterations
    engine.run(range(2), max_epochs=2)

    # ema_model and model can be DP or DDP
    # explicitly cast to float32 to avoid test failure on XLA devices
    ema_weight = _unwrap_model(ema_handler.ema_model).weight.data.to(torch.float32)
    model_weight = _unwrap_model(model).weight.data.to(torch.float32)
    assert ema_weight.device == device
    assert model_weight.device == device
    if interval == 1:
        assert ema_weight.allclose(ema_weight.new_full((1, 2), 4.0625))
    elif interval == 2:
        assert ema_weight.allclose(ema_weight.new_full((1, 2), 3.5))
    else:
        pass
    assert model_weight.allclose(model_weight.new_full((1, 2), 5.0))


@pytest.mark.parametrize("interval", [1, 2])
def test_ema_final_weight_cpu(get_dummy_model, interval):
    device = torch.device("cpu")
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=False, interval=interval)


@pytest.mark.parametrize("interval", [1, 2])
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_ema_final_weight_cuda(get_dummy_model, interval):
    device = torch.device("cuda:0")
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=False, interval=interval)


@pytest.mark.distributed
@pytest.mark.parametrize("interval", [1, 2])
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_ema_final_weight_distrib_nccl_gpu(get_dummy_model, distributed_context_single_node_nccl, interval):
    device = idist.device()
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=True, interval=interval)


@pytest.mark.distributed
@pytest.mark.parametrize("interval", [1, 2])
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_ema_final_weight_distrib_gloo_cpu_or_gpu(get_dummy_model, distributed_context_single_node_gloo, interval):
    device = idist.device()
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=True, interval=interval)


@pytest.mark.distributed
@pytest.mark.parametrize("interval", [1, 2])
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_ema_final_weight_distrib_hvd(get_dummy_model, gloo_hvd_executor, interval):
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()
    # pass device = None to the executor. Different from other distributed tests where the processes are
    # already spawn in the context, the processes here will be explicitly spawn by the executor, so we
    # pass None to the function, and call idist.device() in side the function to get the corresponding device
    gloo_hvd_executor(_test_ema_final_weight, (get_dummy_model(), None, True, interval), np=nproc, do_init=True)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_ema_final_weight_distrib_single_device_xla(get_dummy_model):
    device = idist.device()
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=True)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_ema_final_weight_distrib_xla_nprocs(get_dummy_model, xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])

    def _test_ema_final_weight_xla_nprocs(index):
        device = idist.device()
        _test_ema_final_weight(get_dummy_model(), device=device, ddp=True)

    xmp_executor(_test_ema_final_weight_xla_nprocs, args=(), nprocs=n)


@pytest.mark.multinode_distributed
@pytest.mark.parametrize("interval", [1, 2])
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_ema_final_weight_distrib_multinode_gloo_cpu_or_gpu(
    get_dummy_model, distributed_context_multi_node_gloo, interval
):
    device = idist.device()
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=True, interval=interval)


@pytest.mark.multinode_distributed
@pytest.mark.parametrize("interval", [1, 2])
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_ema_final_weight_distrib_multinode_nccl_gpu(get_dummy_model, distributed_context_multi_node_nccl, interval):
    device = idist.device()
    _test_ema_final_weight(get_dummy_model(), device=device, ddp=True, interval=interval)



from typing import Callable, Optional, Union
from unittest.mock import patch

import pytest
import torch
import torchvision

from ignite.metrics.gan.utils import _BaseInceptionMetric, InceptionModel


class DummyInceptionMetric(_BaseInceptionMetric):
    def __init__(
        self,
        num_features: Optional[int] = None,
        feature_extractor: Optional[torch.nn.Module] = None,
        output_transform: Callable = lambda x: x,
        device: Union[str, torch.device] = torch.device("cpu"),
    ) -> None:
        super(DummyInceptionMetric, self).__init__(
            num_features=num_features,
            feature_extractor=feature_extractor,
            output_transform=output_transform,
            device=device,
        )

    def reset(self):
        pass

    def compute(self):
        pass

    def update(self, output):
        self._extract_features(output)


def test_dummy_metric():
    with pytest.raises(ValueError, match=r"Argument num_features must be greater to zero, got:"):
        DummyInceptionMetric(num_features=-1, feature_extractor=torch.nn.Identity()).update(torch.rand(2, 0))

    with pytest.raises(ValueError, match=r"feature_extractor output must be a tensor of dim 2, got: 1"):
        DummyInceptionMetric(num_features=1000, feature_extractor=torch.nn.Identity()).update(torch.rand(3))

    with pytest.raises(ValueError, match=r"Batch size should be greater than one, got: 0"):
        DummyInceptionMetric(num_features=1000, feature_extractor=torch.nn.Identity()).update(torch.rand(0, 0))

    with pytest.raises(ValueError, match=r"num_features returned by feature_extractor should be 1000, got: 0"):
        DummyInceptionMetric(num_features=1000, feature_extractor=torch.nn.Identity()).update(torch.rand(2, 0))

    with pytest.raises(ValueError, match=r"Argument num_features must be provided, if feature_extractor is specified."):
        DummyInceptionMetric(feature_extractor=torch.nn.Identity())

    with pytest.raises(TypeError, match=r"Argument feature_extractor must be of type torch.nn.Module, got"):
        DummyInceptionMetric(num_features=1000, feature_extractor=lambda x: x)

    assert isinstance(DummyInceptionMetric(num_features=10)._feature_extractor, torch.nn.Identity)


def test_inception_extractor_wrong_inputs():
    with pytest.raises(ValueError, match=r"Inputs should be a tensor of dim 4"):
        InceptionModel(return_features=True)(torch.rand(2))

    with pytest.raises(ValueError, match=r"Inputs should be a tensor with 3 channels"):
        InceptionModel(return_features=True)(torch.rand(2, 2, 2, 0))


def test_inception_model_probability():
    x = torch.rand(2, 3, 299, 299)
    y = InceptionModel(return_features=False)(x)
    assert pytest.approx(torch.sum(y[0]).item()) == 1.0
    assert pytest.approx(torch.sum(y[1]).item()) == 1.0
    assert torch.all(0 <= y)


@pytest.fixture()
def mock_no_torchvision():
    with patch.dict("sys.modules", {"torchvision": None}):
        yield torchvision


def test_no_torchvision(mock_no_torchvision):
    with pytest.raises(ModuleNotFoundError, match=r"This module requires torchvision to be installed."):
        InceptionModel(return_features=True)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_device_mismatch_cuda():
    images = torch.rand(10, 3, 299, 299)
    result = InceptionModel(return_features=False, device="cuda")(images)
    assert result.is_cuda
    assert result.shape == torch.Size([10, 1000])
    result = InceptionModel(return_features=False)(images.cuda())
    assert not result.is_cuda
    assert result.shape == torch.Size([10, 1000])

    images = torch.rand(10, 5)
    result = DummyInceptionMetric(num_features=5, device="cuda")._extract_features(images)
    assert result.is_cuda
    assert result.shape == torch.Size([10, 5])
    result = DummyInceptionMetric(num_features=5)._extract_features(images.cuda())
    assert not result.is_cuda
    assert result.shape == torch.Size([10, 5])



from typing import Tuple

import numpy as np
import pytest
import torch
from torch import Tensor

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics import CosineSimilarity


def test_zero_sample():
    cos_sim = CosineSimilarity()
    with pytest.raises(
        NotComputableError, match=r"CosineSimilarity must have at least one example before it can be computed"
    ):
        cos_sim.compute()


@pytest.fixture(params=list(range(4)))
def test_case(request):
    torch.manual_seed(0)  # For reproducibility

    eps = float(torch.empty(1).uniform_(-8, 0).exp())  # 10 ** uniform(-8, 0)

    return [
        (torch.randn((100, 50)), torch.randn((100, 50)), eps, 1),
        (torch.normal(1.0, 2.0, size=(100, 10)), torch.normal(3.0, 4.0, size=(100, 10)), eps, 1),
        (torch.rand((100, 128)), torch.rand((100, 128)), eps, 16),
        (torch.normal(0.0, 5.0, size=(100, 30)), torch.normal(5.0, 1.0, size=(100, 30)), eps, 16),
    ][request.param]


@pytest.mark.parametrize("n_times", range(5))
def test_compute(n_times, test_case: Tuple[Tensor, Tensor, float, int], available_device):
    y_pred, y, eps, batch_size = test_case

    cos = CosineSimilarity(eps=eps, device=available_device)
    assert cos._device == torch.device(available_device)

    cos.reset()
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            cos.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        cos.update((y_pred, y))

    y_norm = torch.clamp(torch.norm(y, dim=1, keepdim=True), min=eps)
    y_pred_norm = torch.clamp(torch.norm(y_pred, dim=1, keepdim=True), min=eps)

    cosine_sim = torch.sum((y / y_norm) * (y_pred / y_pred_norm), dim=1)
    expected = cosine_sim.mean().item()

    result = cos.compute()

    assert isinstance(result, float)
    assert pytest.approx(expected, rel=2e-5) == result


def test_accumulator_detached(available_device):
    cos = CosineSimilarity(device=available_device)
    assert cos._device == torch.device(available_device)

    y_pred = torch.tensor([[2.0, 3.0], [-2.0, 1.0]], dtype=torch.float)
    y = torch.ones(2, 2, dtype=torch.float)
    cos.update((y_pred, y))

    assert not cos._sum_of_cos_similarities.requires_grad


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_integration(self):
        tol = 2e-5
        n_iters = 100
        batch_size = 10
        n_dims = 100

        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        for metric_device in metric_devices:
            y_true = torch.randn((n_iters * batch_size, n_dims)).float().to(device)
            y_preds = torch.normal(2.0, 3.0, size=(n_iters * batch_size, n_dims)).float().to(device)

            engine = Engine(
                lambda e, i: (
                    y_preds[i * batch_size : (i + 1) * batch_size],
                    y_true[i * batch_size : (i + 1) * batch_size],
                )
            )

            m = CosineSimilarity(device=metric_device)
            m.attach(engine, "cosine_similarity")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=1)

            y_preds = idist.all_gather(y_preds)
            y_true = idist.all_gather(y_true)

            assert "cosine_similarity" in engine.state.metrics
            res = engine.state.metrics["cosine_similarity"]

            y_true_np = y_true.cpu().numpy()
            y_preds_np = y_preds.cpu().numpy()
            y_true_norm = np.clip(np.linalg.norm(y_true_np, axis=1, keepdims=True), 1e-8, None)
            y_preds_norm = np.clip(np.linalg.norm(y_preds_np, axis=1, keepdims=True), 1e-8, None)
            true_res = np.sum((y_true_np / y_true_norm) * (y_preds_np / y_preds_norm), axis=1)
            true_res = np.mean(true_res)

            assert pytest.approx(res, rel=tol) == true_res

    def test_accumulator_device(self):
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)
        for metric_device in metric_devices:
            cos = CosineSimilarity(device=metric_device)

            for dev in (cos._device, cos._sum_of_cos_similarities.device):
                assert dev == metric_device, f"{type(dev)}:{dev} vs {type(metric_device)}:{metric_device}"

            y_pred = torch.tensor([[2.0, 3.0], [-2.0, 1.0]]).float()
            y = torch.ones(2, 2).float()
            cos.update((y_pred, y))

            for dev in (cos._device, cos._sum_of_cos_similarities.device):
                assert dev == metric_device, f"{type(dev)}:{dev} vs {type(metric_device)}:{metric_device}"



from typing import Any, Callable, cast, Tuple, Union

import torch

import ignite.distributed as idist
from ignite.exceptions import NotComputableError
from ignite.metrics.epoch_metric import EpochMetric


def precision_recall_curve_compute_fn(y_preds: torch.Tensor, y_targets: torch.Tensor) -> Tuple[Any, Any, Any]:
    from sklearn.metrics import precision_recall_curve

    y_true = y_targets.cpu().numpy()
    y_pred = y_preds.cpu().numpy()
    return precision_recall_curve(y_true, y_pred)


class PrecisionRecallCurve(EpochMetric):
    """Compute precision-recall pairs for different probability thresholds for binary classification task
    by accumulating predictions and the ground-truth during an epoch and applying
    `sklearn.metrics.precision_recall_curve <https://scikit-learn.org/stable/modules/generated/
    sklearn.metrics.precision_recall_curve.html#sklearn.metrics.precision_recall_curve>`_ .

    Args:
        output_transform: a callable that is used to transform the
            :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
            form expected by the metric. This can be useful if, for example, you have a multi-output model and
            you want to compute the metric with respect to one of the outputs.
        check_compute_fn: Default False. If True, `precision_recall_curve
            <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_curve.html
            #sklearn.metrics.precision_recall_curve>`_ is run on the first batch of data to ensure there are
            no issues. User will be warned in case there are any issues computing the function.
        skip_unrolling: specifies whether output should be unrolled before being fed to update method. Should be
            true for multi-output model, for example, if ``y_pred`` contains multi-ouput as ``(y_pred_a, y_pred_b)``
            Alternatively, ``output_transform`` can be used to handle this.

    Note:
        PrecisionRecallCurve expects y to be comprised of 0's and 1's. y_pred must either be probability estimates
        or confidence values. To apply an activation to y_pred, use output_transform as shown below:

        .. code-block:: python

            def sigmoid_output_transform(output):
                y_pred, y = output
                y_pred = torch.sigmoid(y_pred)
                return y_pred, y
            avg_precision = PrecisionRecallCurve(sigmoid_output_transform)

    Examples:

        .. include:: defaults.rst
            :start-after: :orphan:

        .. testcode::

            y_pred = torch.tensor([0.0474, 0.5987, 0.7109, 0.9997])
            y_true = torch.tensor([0, 0, 1, 1])
            prec_recall_curve = PrecisionRecallCurve()
            prec_recall_curve.attach(default_evaluator, 'prec_recall_curve')
            state = default_evaluator.run([[y_pred, y_true]])

            print("Precision", [round(i, 4) for i in state.metrics['prec_recall_curve'][0].tolist()])
            print("Recall", [round(i, 4) for i in state.metrics['prec_recall_curve'][1].tolist()])
            print("Thresholds", [round(i, 4) for i in state.metrics['prec_recall_curve'][2].tolist()])

        .. testoutput::

            Precision [0.5, 0.6667, 1.0, 1.0, 1.0]
            Recall [1.0, 1.0, 1.0, 0.5, 0.0]
            Thresholds [0.0474, 0.5987, 0.7109, 0.9997]

    .. versionchanged:: 0.5.1
        ``skip_unrolling`` argument is added.
    """

    def __init__(
        self,
        output_transform: Callable = lambda x: x,
        check_compute_fn: bool = False,
        device: Union[str, torch.device] = torch.device("cpu"),
        skip_unrolling: bool = False,
    ) -> None:
        try:
            from sklearn.metrics import precision_recall_curve  # noqa: F401
        except ImportError:
            raise ModuleNotFoundError("This module requires scikit-learn to be installed.")

        super(PrecisionRecallCurve, self).__init__(
            precision_recall_curve_compute_fn,  # type: ignore[arg-type]
            output_transform=output_transform,
            check_compute_fn=check_compute_fn,
            device=device,
            skip_unrolling=skip_unrolling,
        )

    def compute(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:  # type: ignore[override]
        if len(self._predictions) < 1 or len(self._targets) < 1:
            raise NotComputableError("PrecisionRecallCurve must have at least one example before it can be computed.")

        if self._result is None:
            _prediction_tensor = torch.cat(self._predictions, dim=0)
            _target_tensor = torch.cat(self._targets, dim=0)

            ws = idist.get_world_size()
            if ws > 1:
                # All gather across all processes
                _prediction_tensor = cast(torch.Tensor, idist.all_gather(_prediction_tensor))
                _target_tensor = cast(torch.Tensor, idist.all_gather(_target_tensor))

            if idist.get_rank() == 0:
                # Run compute_fn on zero rank only
                precision, recall, thresholds = cast(Tuple, self.compute_fn(_prediction_tensor, _target_tensor))
                precision = torch.tensor(precision, device=_prediction_tensor.device, dtype=self._double_dtype)
                recall = torch.tensor(recall, device=_prediction_tensor.device, dtype=self._double_dtype)
                # thresholds can have negative strides, not compatible with torch tensors
                # https://discuss.pytorch.org/t/negative-strides-in-tensor-error/134287/2
                thresholds = torch.tensor(thresholds.copy(), device=_prediction_tensor.device, dtype=self._double_dtype)
            else:
                precision, recall, thresholds = None, None, None

            if ws > 1:
                # broadcast result to all processes
                precision = idist.broadcast(precision, src=0, safe_mode=True)
                recall = idist.broadcast(recall, src=0, safe_mode=True)
                thresholds = idist.broadcast(thresholds, src=0, safe_mode=True)

            self._result = (precision, recall, thresholds)  # type: ignore[assignment]

        return cast(Tuple[torch.Tensor, torch.Tensor, torch.Tensor], self._result)



# coding: utf-8
import argparse
import random
from collections import OrderedDict
from pathlib import Path

import numpy as np
import torch
import utils
from handlers import Progbar
from torch.optim import Adam
from torch.utils.data import DataLoader
from torchvision import datasets, transforms
from transformer_net import TransformerNet
from vgg import Vgg16

from ignite.engine import Engine, Events
from ignite.handlers import ModelCheckpoint


def check_paths(args):
    try:
        if args.checkpoint_model_dir is not None and not (Path(args.checkpoint_model_dir).exists()):
            Path(args.checkpoint_model_dir).mkdir(parents=True)
    except OSError as e:
        raise OSError(e)


def check_manual_seed(args):
    seed = args.seed or random.randint(1, 10000)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def check_dataset(args):
    transform = transforms.Compose(
        [
            transforms.Resize(args.image_size),
            transforms.CenterCrop(args.image_size),
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.mul(255)),
        ]
    )

    if args.dataset in {"folder", "mscoco"}:
        train_dataset = datasets.ImageFolder(args.dataroot, transform)
    elif args.dataset == "test":
        train_dataset = datasets.FakeData(
            size=args.batch_size, image_size=(3, 32, 32), num_classes=1, transform=transform
        )
    else:
        raise RuntimeError(f"Invalid dataset name: {args.dataset}")

    train_loader = DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True, num_workers=0)

    return train_loader


def train(args):
    device = torch.device("cuda" if args.cuda else "cpu")

    train_loader = check_dataset(args)
    transformer = TransformerNet().to(device)
    optimizer = Adam(transformer.parameters(), args.lr)
    mse_loss = torch.nn.MSELoss()

    vgg = Vgg16(requires_grad=False).to(device)
    style_transform = transforms.Compose([transforms.ToTensor(), transforms.Lambda(lambda x: x.mul(255))])

    style = utils.load_image(args.style_image, size=args.style_size)
    style = style_transform(style)
    style = style.repeat(args.batch_size, 1, 1, 1).to(device)

    features_style = vgg(utils.normalize_batch(style))
    gram_style = [utils.gram_matrix(y) for y in features_style]

    running_avgs = OrderedDict()

    def step(engine, batch):
        x, _ = batch
        x = x.to(device)

        n_batch = len(x)

        optimizer.zero_grad()

        y = transformer(x)

        x = utils.normalize_batch(x)
        y = utils.normalize_batch(y)

        features_x = vgg(x)
        features_y = vgg(y)

        content_loss = args.content_weight * mse_loss(features_y.relu2_2, features_x.relu2_2)

        style_loss = 0.0
        for ft_y, gm_s in zip(features_y, gram_style):
            gm_y = utils.gram_matrix(ft_y)
            style_loss += mse_loss(gm_y, gm_s[:n_batch, :, :])
        style_loss *= args.style_weight

        total_loss = content_loss + style_loss
        total_loss.backward()
        optimizer.step()

        return {"content_loss": content_loss.item(), "style_loss": style_loss.item(), "total_loss": total_loss.item()}

    trainer = Engine(step)
    checkpoint_handler = ModelCheckpoint(
        args.checkpoint_model_dir, "checkpoint", n_saved=10, require_empty=False, create_dir=True
    )
    progress_bar = Progbar(loader=train_loader, metrics=running_avgs)

    trainer.add_event_handler(
        event_name=Events.EPOCH_COMPLETED(every=args.checkpoint_interval),
        handler=checkpoint_handler,
        to_save={"net": transformer},
    )
    trainer.add_event_handler(event_name=Events.ITERATION_COMPLETED, handler=progress_bar)
    trainer.run(train_loader, max_epochs=args.epochs)


def stylize(args):
    device = torch.device("cuda" if args.cuda else "cpu")

    content_transform = transforms.Compose([transforms.ToTensor(), transforms.Lambda(lambda x: x.mul(255))])

    content_image = utils.load_image(args.content_image, scale=args.content_scale)
    content_image = content_transform(content_image)
    content_image = content_image.unsqueeze(0).to(device)

    with torch.no_grad():
        style_model = torch.load(args.model)
        style_model.to(device)
        output = style_model(content_image).cpu()
        utils.save_image(args.output_image, output[0])


def main():
    main_arg_parser = argparse.ArgumentParser(description="parser for fast-neural-style")
    subparsers = main_arg_parser.add_subparsers(title="subcommands", dest="subcommand")

    train_arg_parser = subparsers.add_parser("train", help="parser for training arguments")
    train_arg_parser.add_argument("--epochs", type=int, default=2, help="number of training epochs, default is 2")
    train_arg_parser.add_argument("--batch_size", type=int, default=8, help="batch size for training, default is 8")
    train_arg_parser.add_argument(
        "--dataset", type=str, required=True, choices={"test", "folder", "mscoco"}, help="type of dataset to be used."
    )
    train_arg_parser.add_argument(
        "--dataroot",
        type=str,
        required=True,
        help="path to training dataset, the path should point to a folder "
        "containing another folder with all the training images",
    )
    train_arg_parser.add_argument("--style_image", type=str, default="test", help="path to style-image")
    train_arg_parser.add_argument("--test_image", type=str, default="test", help="path to test-image")
    train_arg_parser.add_argument(
        "--checkpoint_model_dir",
        type=str,
        default="/tmp/checkpoints",
        help="path to folder where checkpoints of trained models will be saved",
    )
    train_arg_parser.add_argument(
        "--checkpoint_interval",
        type=int,
        default=1,
        help="number of batches after which a checkpoint of trained model will be created",
    )
    train_arg_parser.add_argument(
        "--image_size", type=int, default=256, help="size of training images, default is 256 X 256"
    )
    train_arg_parser.add_argument(
        "--style_size", type=int, default=None, help="size of style-image, default is the original size of style image"
    )
    train_arg_parser.add_argument("--cuda", type=int, default=1, help="set it to 1 for running on GPU, 0 for CPU")
    train_arg_parser.add_argument("--seed", type=int, default=42, help="random seed for training")
    train_arg_parser.add_argument(
        "--content_weight", type=float, default=1e5, help="weight for content-loss, default is 1e5"
    )
    train_arg_parser.add_argument(
        "--style_weight", type=float, default=1e10, help="weight for style-loss, default is 1e10"
    )
    train_arg_parser.add_argument("--lr", type=float, default=1e-3, help="learning rate, default is 1e-3")

    eval_arg_parser = subparsers.add_parser("eval", help="parser for evaluation/stylizing arguments")
    eval_arg_parser.add_argument(
        "--content_image", type=str, required=True, help="path to content image you want to stylize"
    )
    eval_arg_parser.add_argument(
        "--content_scale", type=float, default=None, help="factor for scaling down the content image"
    )
    eval_arg_parser.add_argument("--output_image", type=str, required=True, help="path for saving the output image")
    eval_arg_parser.add_argument(
        "--model", type=str, required=True, help="saved model to be used for stylizing the image."
    )
    eval_arg_parser.add_argument("--cuda", type=int, required=True, help="set it to 1 for running on GPU, 0 for CPU")

    args = main_arg_parser.parse_args()

    if args.subcommand is None:
        raise ValueError("ERROR: specify either train or eval")
    if args.cuda and not torch.cuda.is_available():
        raise ValueError("ERROR: cuda is not available, try running on CPU")

    if args.subcommand == "train":
        check_manual_seed(args)
        check_paths(args)
        train(args)
    else:
        stylize(args)


if __name__ == "__main__":
    main()



import os

import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import _InfiniteConstantSampler
from torch.utils.data.dataset import Dataset, IterableDataset
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data.sampler import BatchSampler, RandomSampler, Sampler, SequentialSampler, WeightedRandomSampler

import ignite.distributed as idist
from ignite.distributed.auto import auto_dataloader, auto_model, auto_optim, DistributedProxySampler
from tests.ignite import is_mps_available_and_functional


class DummyDS(Dataset):
    def __init__(self, length=10):
        self.length = length

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        return index


class DummyIterableDataset(IterableDataset):
    def __init__(self, start, end):
        super(DummyIterableDataset).__init__()
        self.start = start
        self.end = end

    def __iter__(self):
        return iter(range(self.start, self.end))

    def __len__(self):
        return self.end - self.start


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("WORLD_SIZE" not in os.environ, reason="Skip if WORLD_SIZE not in env vars")
def test_auto_dataloader_warning(distributed_context_single_node_gloo):
    with pytest.warns(UserWarning, match=r"Found batch_sampler in provided kwargs"):
        auto_dataloader(
            DummyDS(), batch_sampler=BatchSampler(SequentialSampler(range(10)), batch_size=3, drop_last=False)
        )


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("WORLD_SIZE" not in os.environ, reason="Skip if WORLD_SIZE not in env vars")
def test_auto_dataloader_warning_distributed_sampler(distributed_context_single_node_gloo):
    dataset = DummyDS()
    rank = idist.get_rank()
    world_size = idist.get_world_size()
    auto_dataloader(dataset, sampler=DistributedSampler(dataset, num_replicas=world_size, rank=rank))
    if world_size > 1:
        wrong_rank = (rank + 1) % world_size
        expected_warning = f"Found distributed sampler with rank={wrong_rank}, but process rank is {rank}"
        with pytest.warns(UserWarning, match=expected_warning):
            auto_dataloader(dataset, sampler=DistributedSampler(dataset, num_replicas=world_size, rank=wrong_rank))
    expected_warning = f"Found distributed sampler with num_replicas={world_size + 1}, but world size is {world_size}"
    with pytest.warns(UserWarning, match=expected_warning):
        auto_dataloader(dataset, sampler=DistributedSampler(dataset, num_replicas=world_size + 1, rank=rank))


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_auto_dataloader_warning_tpu():
    with pytest.warns(UserWarning, match=r"Found incompatible options: xla support and pin_memory"):
        auto_dataloader(DummyDS(), pin_memory=True)


def _test_auto_dataloader(ws, nproc, batch_size, num_workers=1, sampler_name=None, dl_type=DataLoader):
    def _test(data):
        if sampler_name is None:
            sampler = None
        elif sampler_name == "WeightedRandomSampler":
            sampler = WeightedRandomSampler(weights=torch.ones(100), num_samples=100)
        elif sampler_name == "DistributedSampler":
            sampler = DistributedSampler(data, num_replicas=ws, rank=idist.get_rank())
        else:
            raise RuntimeError(f"Unknown sampler name: {sampler_name}")

        # Test auto_dataloader
        assert idist.get_world_size() == ws, f"{idist.get_world_size()} vs {ws}"

        shuffle = sampler is None if not isinstance(data, IterableDataset) else False
        dataloader = auto_dataloader(
            data, batch_size=batch_size, num_workers=num_workers, sampler=sampler, shuffle=shuffle
        )

        assert isinstance(dataloader, dl_type)
        if hasattr(dataloader, "_loader"):
            dataloader = dataloader._loader
        if ws < batch_size:
            assert dataloader.batch_size == batch_size // ws
        else:
            assert dataloader.batch_size == batch_size
        if ws <= num_workers:
            assert dataloader.num_workers == (num_workers + nproc - 1) // nproc
        else:
            assert dataloader.num_workers == num_workers

        if isinstance(data, IterableDataset):
            sampler_type = _InfiniteConstantSampler
        elif ws > 1:
            if sampler is None or isinstance(sampler, DistributedSampler):
                sampler_type = DistributedSampler
            else:
                sampler_type = DistributedProxySampler
        else:
            sampler_type = RandomSampler if sampler is None else type(sampler)

        assert isinstance(dataloader.sampler, sampler_type)

        if isinstance(dataloader, DataLoader):
            assert dataloader.pin_memory == ("cuda" in idist.device().type)

    data = torch.rand(100, 3, 12, 12)
    _test(data)
    if sampler_name is None:
        data = DummyIterableDataset(0, 100)
        _test(data)


def _test_auto_model(model, ws, device, sync_bn=False, **kwargs):
    model = auto_model(model, sync_bn=sync_bn, **kwargs)
    bnd = idist.backend()
    if ws > 1 and torch.device(device).type in ("cuda", "cpu"):
        if idist.has_native_dist_support and bnd in ("nccl", "gloo"):
            assert isinstance(model, nn.parallel.DistributedDataParallel)
            if sync_bn:
                assert any([isinstance(m, nn.SyncBatchNorm) for m in model.modules()])
            if "find_unused_parameters" in kwargs:
                assert model.find_unused_parameters == kwargs["find_unused_parameters"]
        elif idist.has_hvd_support and bnd in ("horovod",):
            assert isinstance(model, nn.Module)
    elif device != "cpu" and torch.cuda.is_available() and torch.cuda.device_count() > 1:
        assert isinstance(model, nn.parallel.DataParallel)
    else:
        assert isinstance(model, nn.Module)

    assert all(
        [p.device.type == torch.device(device).type for p in model.parameters()]
    ), f"{[p.device.type for p in model.parameters()]} vs {torch.device(device).type}"


def _test_auto_model_optimizer(ws, device):
    # Test auto_model
    model = nn.Linear(10, 10)
    _test_auto_model(model, ws, device)

    model = nn.Sequential(nn.Linear(20, 100), nn.BatchNorm1d(100))
    _test_auto_model(model, ws, device, sync_bn="cuda" in torch.device(device).type)
    if ws > 1:
        _test_auto_model(model, ws, device, find_unused_parameters=True)
        _test_auto_model(model, ws, device, find_unused_parameters=False)

    # Test auto_optim
    bnd = idist.backend()
    optimizer = optim.SGD(model.parameters(), lr=0.01)
    optimizer = auto_optim(optimizer)
    if idist.has_xla_support and "xla" in device:
        assert isinstance(optimizer, optim.SGD) and hasattr(optimizer, "wrapped_optimizer")
    elif idist.has_hvd_support and bnd in ("horovod",):
        assert isinstance(optimizer, optim.SGD) and hasattr(optimizer, "_allreduce_grad_async")
    else:
        assert isinstance(optimizer, optim.SGD) and not hasattr(optimizer, "wrapped_optimizer")

    if idist.has_hvd_support and bnd in ("horovod",):
        backward_passes_per_step = 2
        optimizer = optim.SGD(model.parameters(), lr=0.01)
        optimizer = auto_optim(optimizer, backward_passes_per_step=backward_passes_per_step)
        assert isinstance(optimizer, optim.SGD) and hasattr(optimizer, "backward_passes_per_step")
        assert optimizer.backward_passes_per_step == backward_passes_per_step


@pytest.mark.skipif(not is_mps_available_and_functional(), reason="Skip if MPS not functional")
def test_auto_methods_no_dist():
    _test_auto_dataloader(1, 1, batch_size=1)
    _test_auto_dataloader(1, 1, batch_size=10, num_workers=2)
    _test_auto_dataloader(1, 1, batch_size=10, sampler_name="WeightedRandomSampler")
    _test_auto_dataloader(1, 1, batch_size=10, sampler_name="DistributedSampler")
    device = idist.device()
    _test_auto_model_optimizer(1, device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_auto_methods_gloo(distributed_context_single_node_gloo):
    ws = distributed_context_single_node_gloo["world_size"]
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=10, num_workers=2)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=10, sampler_name="WeightedRandomSampler")
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=10, sampler_name="DistributedSampler")

    device = idist.device()
    _test_auto_model_optimizer(ws, device)

    if ws > 1 and device.type == "cpu":
        # Pytorch <= 1.9.0 => AssertionError
        # Pytorch >  1.9   => ValueError
        # https://github.com/pytorch/pytorch/blob/master/torch/nn/parallel/distributed.py#L1498
        with pytest.raises(
            (AssertionError, ValueError), match=r"SyncBatchNorm layers only work with (GPU|CUDA) modules"
        ):
            model = nn.Sequential(nn.Linear(20, 100), nn.BatchNorm1d(100))
            auto_model(model, sync_bn=True)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_auto_methods_nccl(distributed_context_single_node_nccl):
    ws = distributed_context_single_node_nccl["world_size"]
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=10, num_workers=10)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1, sampler_name="WeightedRandomSampler")
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1, sampler_name="DistributedSampler")

    device = idist.device()
    _test_auto_model_optimizer(ws, device)

    if ws > 1:
        with pytest.raises(ValueError, match=r"Argument kwargs should not contain 'device_ids'"):
            auto_model(nn.Linear(1, 1), device_ids=[0])


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_auto_methods_hvd(gloo_hvd_executor):
    device = "cpu" if not torch.cuda.is_available() else "cuda"
    np = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_auto_dataloader, args=(np, np, 1), np=np, do_init=True)
    gloo_hvd_executor(_test_auto_dataloader, args=(np, np, 10, 10), np=np, do_init=True)
    gloo_hvd_executor(_test_auto_dataloader, args=(np, np, 1, 1, "WeightedRandomSampler"), np=np, do_init=True)
    gloo_hvd_executor(_test_auto_dataloader, args=(np, np, 1, 1, "DistributedSampler"), np=np, do_init=True)

    gloo_hvd_executor(_test_auto_model_optimizer, args=(np, device), np=np, do_init=True)


def _test_auto_methods_xla(index, ws):
    dl_type = DataLoader
    if ws > 1:
        from ignite.distributed.auto import _MpDeviceLoader

        dl_type = _MpDeviceLoader
        try:
            from torch_xla.distributed.parallel_loader import MpDeviceLoader

            dl_type = MpDeviceLoader
        except ImportError:
            pass

    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1, dl_type=dl_type)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=10, num_workers=2, dl_type=dl_type)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1, sampler_name="WeightedRandomSampler", dl_type=dl_type)
    _test_auto_dataloader(ws=ws, nproc=ws, batch_size=1, sampler_name="DistributedSampler", dl_type=dl_type)

    device = "xla"
    _test_auto_model_optimizer(ws, device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_auto_methods_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_auto_methods_xla, args=(n,), nprocs=n)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_auto_methods_xla():
    _test_auto_methods_xla(index=0, ws=1)


def test_dist_proxy_sampler():
    weights = torch.ones(100)
    weights[:50] += 1
    num_samples = 200
    sampler = WeightedRandomSampler(weights, num_samples)

    num_replicas = 8
    dist_samplers = [DistributedProxySampler(sampler, num_replicas=num_replicas, rank=i) for i in range(num_replicas)]

    for seed in range(100):
        torch.manual_seed(seed)
        true_indices = list(sampler)

        indices_per_rank = []
        for s in dist_samplers:
            s.set_epoch(seed)
            indices_per_rank += list(s)

        set_indices_per_rank = set(indices_per_rank)
        set_true_indices = set(true_indices)
        assert (
            set_indices_per_rank == set_true_indices
        ), f"{set_true_indices - set_indices_per_rank} | {set_indices_per_rank - set_true_indices}"

    with pytest.raises(TypeError, match=r"Argument sampler should be instance of torch Sampler"):
        DistributedProxySampler(None)

    with pytest.raises(TypeError, match=r"Argument sampler should have length"):
        DistributedProxySampler(Sampler([1]))

    with pytest.raises(TypeError, match=r"Argument sampler must not be a distributed sampler already"):
        DistributedProxySampler(DistributedSampler(sampler, num_replicas=num_replicas, rank=0))



from PIL import Image


def load_image(filename, size=None, scale=None):
    img = Image.open(filename)
    if size is not None:
        img = img.resize((size, size), Image.LANCZOS)
    elif scale is not None:
        img = img.resize((int(img.size[0] / scale), int(img.size[1] / scale)), Image.LANCZOS)
    return img


def save_image(filename, data):
    img = data.clone().clamp(0, 255).numpy()
    img = img.transpose(1, 2, 0).astype("uint8")
    img = Image.fromarray(img)
    img.save(filename)


def gram_matrix(y):
    (b, ch, h, w) = y.size()
    features = y.view(b, ch, w * h)
    features_t = features.transpose(1, 2)
    gram = features.bmm(features_t) / (ch * h * w)
    return gram


def normalize_batch(batch):
    # normalize using imagenet mean and std
    mean = batch.new_tensor([0.485, 0.456, 0.406]).view(-1, 1, 1)
    std = batch.new_tensor([0.229, 0.224, 0.225]).view(-1, 1, 1)
    batch = batch.div_(255.0)
    return (batch - mean) / std



import torch

try:
    from torch.utils.data import IterableDataset
except ImportError:

    class IterableDataset:
        pass


class BatchChecker:
    def __init__(self, data, init_counter=0):
        self.counter = init_counter
        self.data = data
        self.true_batch = None

    def check(self, batch):
        self.true_batch = self.data[self.counter % len(self.data)]
        self.counter += 1
        res = self.true_batch == batch
        return res.all() if not isinstance(res, bool) else res


class IterationCounter:
    def __init__(self, start_value=1):
        self.current_iteration_count = start_value

    def __call__(self, engine):
        assert engine.state.iteration == self.current_iteration_count
        self.current_iteration_count += 1


class EpochCounter:
    def __init__(self, start_value=1):
        self.current_epoch_count = start_value

    def __call__(self, engine):
        assert engine.state.epoch == self.current_epoch_count
        self.current_epoch_count += 1


def setup_sampler(sampler_type, num_iters, batch_size):
    if sampler_type is None:
        return None, batch_size

    if sampler_type == "weighted":
        from torch.utils.data.sampler import WeightedRandomSampler

        w = torch.ones(num_iters * batch_size, dtype=torch.float)
        for i in range(num_iters):
            w[batch_size * i : batch_size * (i + 1)] += i * 1.0
        return WeightedRandomSampler(w, num_samples=num_iters * batch_size, replacement=True), batch_size

    if sampler_type == "distributed":
        import torch.distributed as dist
        from torch.utils.data.distributed import DistributedSampler

        num_replicas = 1
        rank = 0
        if dist.is_available() and dist.is_initialized():
            num_replicas = dist.get_world_size()
            rank = dist.get_rank()

        dataset = torch.zeros(num_iters * batch_size)
        return DistributedSampler(dataset, num_replicas=num_replicas, rank=rank), batch_size // num_replicas


class MyIterableDataset(IterableDataset):
    def __init__(self, start, end):
        super(MyIterableDataset).__init__()
        assert end > start, "this example code only works with end >= start"
        self.start = start
        self.end = end

    def __iter__(self):
        return iter(range(self.start, self.end))


def get_iterable_dataset(*args, **kwargs):
    return MyIterableDataset(*args, **kwargs)



import sys
from unittest.mock import call, MagicMock

import pytest
import torch

from ignite.engine import Engine, Events, State

from ignite.handlers.mlflow_logger import global_step_from_engine, MLflowLogger, OptimizerParamsHandler, OutputHandler


def test_output_handler_with_wrong_logger_type():
    wrapper = OutputHandler("tag", output_transform=lambda x: x)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(TypeError, match="Handler 'OutputHandler' works only with MLflowLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_output_handler_output_transform():
    wrapper = OutputHandler("tag", output_transform=lambda x: x)
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.output = 12345
    mock_engine.state.iteration = 123

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    mock_logger.log_metrics.assert_called_once_with({"tag output": 12345}, step=123)

    wrapper = OutputHandler("another_tag", output_transform=lambda x: {"loss": x})
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)
    mock_logger.log_metrics.assert_called_once_with({"another_tag loss": 12345}, step=123)


def test_output_handler_metric_names():
    wrapper = OutputHandler("tag", metric_names=["a", "b", "c"])
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45, "c": torch.tensor(10.0)})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.log_metrics.call_count == 1
    mock_logger.log_metrics.assert_called_once_with({"tag a": 12.23, "tag b": 23.45, "tag c": 10.0}, step=5)

    wrapper = OutputHandler("tag", metric_names=["a"])

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": torch.tensor([0.0, 1.0, 2.0, 3.0])})
    mock_engine.state.iteration = 5

    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.log_metrics.call_count == 1
    mock_logger.log_metrics.assert_has_calls(
        [call({"tag a 0": 0.0, "tag a 1": 1.0, "tag a 2": 2.0, "tag a 3": 3.0}, step=5)], any_order=True
    )

    wrapper = OutputHandler("tag", metric_names=["a", "c"])

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 55.56, "c": "Some text"})
    mock_engine.state.iteration = 7

    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    with pytest.warns(UserWarning):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.log_metrics.call_count == 1
    mock_logger.log_metrics.assert_has_calls([call({"tag a": 55.56}, step=7)], any_order=True)

    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(
        metrics={
            "a": 123,
            "b": {"c": [2.34, {"d": 1}]},
            "c": (22, [33, -5.5], {"e": 32.1}),
        }
    )
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.log_metrics.call_count == 1
    mock_logger.log_metrics.assert_called_once_with(
        {
            "tag a": 123,
            "tag b c 0": 2.34,
            "tag b c 1 d": 1,
            "tag c 0": 22,
            "tag c 1 0": 33,
            "tag c 1 1": -5.5,
            "tag c 2 e": 32.1,
        },
        step=5,
    )


def test_output_handler_both():
    wrapper = OutputHandler("tag", metric_names=["a", "b"], output_transform=lambda x: {"loss": x})
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    assert mock_logger.log_metrics.call_count == 1
    mock_logger.log_metrics.assert_called_once_with({"tag a": 12.23, "tag b": 23.45, "tag loss": 12345}, step=5)


def test_output_handler_with_wrong_global_step_transform_output():
    def global_step_transform(*args, **kwargs):
        return "a"

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    with pytest.raises(TypeError, match="global_step must be int"):
        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)


def test_output_handler_with_global_step_transform():
    def global_step_transform(*args, **kwargs):
        return 10

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    mock_logger.log_metrics.assert_called_once_with({"tag loss": 12345}, step=10)


def test_output_handler_with_global_step_from_engine():
    mock_another_engine = MagicMock()
    mock_another_engine.state = State()
    mock_another_engine.state.epoch = 10
    mock_another_engine.state.output = 12.345

    wrapper = OutputHandler(
        "tag",
        output_transform=lambda x: {"loss": x},
        global_step_transform=global_step_from_engine(mock_another_engine),
    )

    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 1
    mock_engine.state.output = 0.123

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.log_metrics.call_count == 1
    mock_logger.log_metrics.assert_has_calls(
        [call({"tag loss": mock_engine.state.output}, step=mock_another_engine.state.epoch)]
    )

    mock_another_engine.state.epoch = 11
    mock_engine.state.output = 1.123

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.log_metrics.call_count == 2
    mock_logger.log_metrics.assert_has_calls(
        [call({"tag loss": mock_engine.state.output}, step=mock_another_engine.state.epoch)]
    )


def test_output_handler_state_attrs():
    wrapper = OutputHandler("tag", state_attributes=["alpha", "beta", "gamma"])
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 5
    mock_engine.state.alpha = 3.899
    mock_engine.state.beta = torch.tensor(12.21)
    mock_engine.state.gamma = torch.tensor([21.0, 6.0])

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    mock_logger.log_metrics.assert_called_once_with(
        {"tag alpha": 3.899, "tag beta": torch.tensor(12.21).item(), "tag gamma 0": 21.0, "tag gamma 1": 6.0}, step=5
    )


def test_optimizer_params_handler_wrong_setup():
    with pytest.raises(TypeError):
        OptimizerParamsHandler(optimizer=None)

    optimizer = MagicMock(spec=torch.optim.Optimizer)
    handler = OptimizerParamsHandler(optimizer=optimizer)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(TypeError, match="Handler OptimizerParamsHandler works only with MLflowLogger"):
        handler(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_optimizer_params():
    optimizer = torch.optim.SGD([torch.tensor(0.0)], lr=0.01)
    wrapper = OptimizerParamsHandler(optimizer=optimizer, param_name="lr")
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 123

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)
    mock_logger.log_metrics.assert_called_once_with({"lr group_0": 0.01}, step=123)

    wrapper = OptimizerParamsHandler(optimizer, param_name="lr", tag="generator")
    mock_logger = MagicMock(spec=MLflowLogger)
    mock_logger.log_metrics = MagicMock()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)
    mock_logger.log_metrics.assert_called_once_with({"generator lr group_0": 0.01}, step=123)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_integration(dirname):
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    trainer = Engine(update_fn)

    mlflow_logger = MLflowLogger(tracking_uri=str(dirname / "mlruns"))

    true_values = []

    def dummy_handler(engine, logger, event_name):
        global_step = engine.state.get_event_attrib_value(event_name)
        v = global_step * 0.1
        true_values.append(v)
        logger.log_metrics({"test_value": v}, step=global_step)

    mlflow_logger.attach(trainer, log_handler=dummy_handler, event_name=Events.EPOCH_COMPLETED)

    import mlflow

    active_run = mlflow.active_run()

    trainer.run(data, max_epochs=n_epochs)
    mlflow_logger.close()

    from mlflow.tracking import MlflowClient

    client = MlflowClient(tracking_uri=str(dirname / "mlruns"))
    stored_values = client.get_metric_history(active_run.info.run_id, "test_value")

    for t, s in zip(true_values, stored_values):
        assert pytest.approx(t) == s.value


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_integration_as_context_manager(dirname):
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    true_values = []

    with MLflowLogger(str(dirname / "mlruns")) as mlflow_logger:
        trainer = Engine(update_fn)

        def dummy_handler(engine, logger, event_name):
            global_step = engine.state.get_event_attrib_value(event_name)
            v = global_step * 0.1
            true_values.append(v)
            logger.log_metrics({"test_value": v}, step=global_step)

        mlflow_logger.attach(trainer, log_handler=dummy_handler, event_name=Events.EPOCH_COMPLETED)

        import mlflow

        active_run = mlflow.active_run()

        trainer.run(data, max_epochs=n_epochs)

    from mlflow.tracking import MlflowClient

    client = MlflowClient(tracking_uri=str(dirname / "mlruns"))
    stored_values = client.get_metric_history(active_run.info.run_id, "test_value")

    for t, s in zip(true_values, stored_values):
        assert pytest.approx(t) == s.value


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_mlflow_bad_metric_name_handling(dirname):
    import mlflow

    true_values = [123.0, 23.4, 333.4]
    with MLflowLogger(str(dirname / "mlruns")) as mlflow_logger:
        active_run = mlflow.active_run()

        handler = OutputHandler(tag="training", metric_names="all")
        engine = Engine(lambda e, b: None)
        engine.state = State(metrics={"metric:0 in %": 123.0, "metric 0": 1000.0})

        with pytest.warns(UserWarning, match=r"MLflowLogger output_handler encountered an invalid metric name"):
            engine.state.epoch = 1
            handler(engine, mlflow_logger, event_name=Events.EPOCH_COMPLETED)

            for _, v in enumerate(true_values):
                engine.state.epoch += 1
                engine.state.metrics["metric 0"] = v
                handler(engine, mlflow_logger, event_name=Events.EPOCH_COMPLETED)

    from mlflow.tracking import MlflowClient

    client = MlflowClient(tracking_uri=str(dirname / "mlruns"))
    stored_values = client.get_metric_history(active_run.info.run_id, "training metric 0")

    for t, s in zip([1000.0] + true_values, stored_values):
        assert t == s.value


@pytest.mark.parametrize("no_site_packages", ["mlflow"], indirect=True)
def test_no_mlflow_client(no_site_packages):
    with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires mlflow to be installed."):
        MLflowLogger()
